READ_DOCSTRING = """
    Read the input ``table`` and return the table.  Most of
    the default behavior for various parameters is determined by the Reader
    class.

    See also:

    - https://docs.astropy.org/en/stable/io/ascii/
    - https://docs.astropy.org/en/stable/io/ascii/read.html

    Parameters
    ----------
    table : str, file-like, list, `pathlib.Path` object
        Input table as a file name, file-like object, list of string[s],
        single newline-separated string or `pathlib.Path` object.
    guess : bool
        Try to guess the table format. Defaults to None.
    format : str, `~astropy.io.ascii.BaseReader`
        Input table format
    Inputter : `~astropy.io.ascii.BaseInputter`
        Inputter class
    Outputter : `~astropy.io.ascii.BaseOutputter`
        Outputter class
    delimiter : str
        Column delimiter string
    comment : str
        Regular expression defining a comment line in table
    quotechar : str
        One-character string to quote fields containing special characters
    header_start : int
        Line index for the header line not counting comment or blank lines.
        A line with only whitespace is considered blank.
    data_start : int
        Line index for the start of data not counting comment or blank lines.
        A line with only whitespace is considered blank.
    data_end : int
        Line index for the end of data not counting comment or blank lines.
        This value can be negative to count from the end.
    converters : dict
        Dictionary of converters. Keys in the dictionary are columns names,
        values are converter functions. In addition to single column names
        you can use wildcards via `fnmatch` to select multiple columns.
    data_Splitter : `~astropy.io.ascii.BaseSplitter`
        Splitter class to split data columns
    header_Splitter : `~astropy.io.ascii.BaseSplitter`
        Splitter class to split header columns
    names : list
        List of names corresponding to each data column
    include_names : list
        List of names to include in output.
    exclude_names : list
        List of names to exclude from output (applied after ``include_names``)
    fill_values : tuple, list of tuple
        specification of fill values for bad or missing table values
    fill_include_names : list
        List of names to include in fill_values.
    fill_exclude_names : list
        List of names to exclude from fill_values (applied after ``fill_include_names``)
    fast_reader : bool, str or dict
        Whether to use the C engine, can also be a dict with options which
        defaults to `False`; parameters for options dict:

        use_fast_converter: bool
            enable faster but slightly imprecise floating point conversion method
        parallel: bool or int
            multiprocessing conversion using ``cpu_count()`` or ``'number'`` processes
        exponent_style: str
            One-character string defining the exponent or ``'Fortran'`` to auto-detect
            Fortran-style scientific notation like ``'3.14159D+00'`` (``'E'``, ``'D'``, ``'Q'``),
            all case-insensitive; default ``'E'``, all other imply ``use_fast_converter``
        chunk_size : int
            If supplied with a value > 0 then read the table in chunks of
            approximately ``chunk_size`` bytes. Default is reading table in one pass.
        chunk_generator : bool
            If True and ``chunk_size > 0`` then return an iterator that returns a
            table for each chunk.  The default is to return a single stacked table
            for all the chunks.

    encoding : str
        Allow to specify encoding to read the file (default= ``None``).

    Returns
    -------
    dat : `~astropy.table.Table` or <generator>
        Output table

    """

# Specify allowed types for core write() keyword arguments.  Each entry
# corresponds to the name of an argument and either a type (e.g. int) or a
# list of types.  These get used in io.ascii.ui._validate_read_write_kwargs().
# -  The commented-out kwargs are too flexible for a useful check
# -  'list-list' is a special case for an iterable that is not a string.
READ_KWARG_TYPES = {
    # 'table'
    'guess': bool,
    # 'format'
    # 'Reader'
    # 'Inputter'
    # 'Outputter'
    'delimiter': str,
    'comment': str,
    'quotechar': str,
    'header_start': int,
    'data_start': (int, str),  # CDS allows 'guess'
    'data_end': int,
    'converters': dict,
    # 'data_Splitter'
    # 'header_Splitter'
    'names': 'list-like',
    'include_names': 'list-like',
    'exclude_names': 'list-like',
    'fill_values': 'list-like',
    'fill_include_names': 'list-like',
    'fill_exclude_names': 'list-like',
    'fast_reader': (bool, str, dict),
    'encoding': str,
}


WRITE_DOCSTRING = """
    Write the input ``table`` to ``filename``.  Most of the default behavior
    for various parameters is determined by the Writer class.

    See also:

    - https://docs.astropy.org/en/stable/io/ascii/
    - https://docs.astropy.org/en/stable/io/ascii/write.html

    Parameters
    ----------
    table : `~astropy.io.ascii.BaseReader`, array-like, str, file-like, list
        Input table as a Reader object, Numpy struct array, file name,
        file-like object, list of strings, or single newline-separated string.
    output : str, file-like
        Output [filename, file-like object]. Defaults to``sys.stdout``.
    format : str
        Output table format. Defaults to 'basic'.
    delimiter : str
        Column delimiter string
    comment : str, bool
        String defining a comment line in table.  If `False` then comments
        are not written out.
    quotechar : str
        One-character string to quote fields containing special characters
    formats : dict
        Dictionary of format specifiers or formatting functions
    strip_whitespace : bool
        Strip surrounding whitespace from column values.
    names : list
        List of names corresponding to each data column
    include_names : list
        List of names to include in output.
    exclude_names : list
        List of names to exclude from output (applied after ``include_names``)
    fast_writer : bool, str
        Whether to use the fast Cython writer.  Can be `True` (use fast writer
        if available), `False` (do not use fast writer), or ``'force'`` (use
        fast writer and fail if not available, mostly for testing).
    overwrite : bool
        If ``overwrite=False`` (default) and the file exists, then an OSError
        is raised. This parameter is ignored when the ``output`` arg is not a
        string (e.g., a file object).

    """
# Specify allowed types for core write() keyword arguments.  Each entry
# corresponds to the name of an argument and either a type (e.g. int) or a
# list of types.  These get used in io.ascii.ui._validate_read_write_kwargs().
# -  The commented-out kwargs are too flexible for a useful check
# -  'list-list' is a special case for an iterable that is not a string.
WRITE_KWARG_TYPES = {
    # 'table'
    # 'output'
    'format': str,
    'delimiter': str,
    'comment': (str, bool),
    'quotechar': str,
    'header_start': int,
    'formats': dict,
    'strip_whitespace': (bool),
    'names': 'list-like',
    'include_names': 'list-like',
    'exclude_names': 'list-like',
    'fast_writer': (bool, str),
    'overwrite': (bool),
}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
This subpackage contains implementations of command-line scripts that are
included with Astropy.

The actual scripts that are installed in bin/ are simple wrappers for these
modules that will run in any Python version.
"""



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""This module contains dictionaries with sets of parameters for a
given cosmology.

Each cosmology has the following parameters defined:

    ==========  =====================================
    Oc0         Omega cold dark matter at z=0
    Ob0         Omega baryon at z=0
    Om0         Omega matter at z=0
    flat        Is this assumed flat?  If not, Ode0 must be specified
    Ode0        Omega dark energy at z=0 if flat is False
    H0          Hubble parameter at z=0 in km/s/Mpc
    n           Density perturbation spectral index
    Tcmb0       Current temperature of the CMB
    Neff        Effective number of neutrino species
    m_nu        Assumed mass of neutrino species, in eV.
    sigma8      Density perturbation amplitude
    tau         Ionisation optical depth
    z_reion     Redshift of hydrogen reionisation
    t0          Age of the universe in Gyr
    reference   Reference for the parameters
    ==========  =====================================

The list of cosmologies available are given by the tuple
`available`. Current cosmologies available:

Planck 2018 (Planck18) parameters from Planck Collaboration 2020,
 A&A, 641, A6 (Paper VI), Table 2 (TT, TE, EE + lowE + lensing + BAO)

Planck 2018 (Planck18_arXiv_v2) parameters from Planck Collaboration 2018,
 arXiv:1807.06209v2 (Paper VI), Table 2 (TT, TE, EE + lowE + lensing + BAO)

Planck 2015 (Planck15) parameters from Planck Collaboration 2016, A&A, 594, A13
 (Paper XIII), Table 4 (TT, TE, EE + lowP + lensing + ext)

Planck 2013 (Planck13) parameters from Planck Collaboration 2014, A&A, 571, A16
 (Paper XVI), Table 5 (Planck + WP + highL + BAO)

WMAP 9 year (WMAP9) parameters from Hinshaw et al. 2013, ApJS, 208, 19,
doi: 10.1088/0067-0049/208/2/19. Table 4 (WMAP9 + eCMB + BAO + H0)

WMAP 7 year (WMAP7) parameters from Komatsu et al. 2011, ApJS, 192, 18,
doi: 10.1088/0067-0049/192/2/18. Table 1 (WMAP + BAO + H0 ML).

WMAP 5 year (WMAP5) parameters from Komatsu et al. 2009, ApJS, 180, 330,
doi: 10.1088/0067-0049/180/2/330. Table 1 (WMAP + BAO + SN ML).

WMAP 3 year (WMAP3) parameters from Spergel et al. 2007, ApJS, 170, 377,
doi:  10.1086/513700. Table 6. (WMAP + SNGold) Obtained from https://lambda.gsfc.nasa.gov/product/map/dr2/params/lcdm_wmap_sngold.cfm
Tcmb0 and Neff are the standard values as also used for WMAP5, 7, 9.
Pending WMAP team approval and subject to change.

WMAP 1 year (WMAP1) parameters from Spergel et al. 2003, ApJS, 148, 175,
doi:  10.1086/377226. Table 7 (WMAP + CBI + ACBAR + 2dFGRS + Lya)
Tcmb0 and Neff are the standard values as also used for WMAP5, 7, 9.
Pending WMAP team approval and subject to change.

"""

# STDLIB
from types import MappingProxyType

# LOCAL
import astropy.units as u

# Note: if you add a new cosmology, please also update the table
# in the 'Built-in Cosmologies' section of astropy/docs/cosmology/index.rst
# in addition to the list above.  You also need to add them to the 'available'
# list at the bottom of this file.

# Planck 2018 paper VI
# Unlike Planck 2015, the paper includes massive neutrinos in Om0, which here
# are included in m_nu.  Hence, the Om0 value differs slightly from the paper.
Planck18 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.2607,
    Ob0=0.04897,
    Om0=0.30966,
    H0=67.66 * (u.km / u.s / u.Mpc),
    n=0.9665,
    sigma8=0.8102,
    tau=0.0561,
    z_reion=7.82,
    t0=13.787 * u.Gyr,
    Tcmb0=2.7255 * u.K,
    Neff=3.046,
    flat=True,
    m_nu=[0., 0., 0.06] * u.eV,
    reference=("Planck Collaboration 2018, 2020, A&A, 641, A6  (Paper VI),"
               " Table 2 (TT, TE, EE + lowE + lensing + BAO)")
))

# Planck 2018 paper VI v2.  Identical to Planck18 above.
# Warning: deprecated and will be removed in future versions.
Planck18_arXiv_v2 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.2607,
    Ob0=0.04897,
    Om0=0.30966,
    H0=67.66 * (u.km / u.s / u.Mpc),
    n=0.9665,
    sigma8=0.8102,
    tau=0.0561,
    z_reion=7.82,
    t0=13.787 * u.Gyr,
    Tcmb0=2.7255 * u.K,
    Neff=3.046,
    flat=True,
    m_nu=[0., 0., 0.06] * u.eV,
    reference=("DEPRECATED: Planck Collaboration 2018, arXiv:1807.06209 v2 (Paper VI),"
               " Table 2 (TT, TE, EE + lowE + lensing + BAO)")
))

# Planck 2015 paper XII Table 4 final column (best fit)
Planck15 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.2589,
    Ob0=0.04860,
    Om0=0.3075,
    H0=67.74 * (u.km / u.s / u.Mpc),
    n=0.9667,
    sigma8=0.8159,
    tau=0.066,
    z_reion=8.8,
    t0=13.799 * u.Gyr,
    Tcmb0=2.7255 * u.K,
    Neff=3.046,
    flat=True,
    m_nu=[0., 0., 0.06] * u.eV,
    reference=("Planck Collaboration 2016, A&A, 594, A13 (Paper XIII),"
               " Table 4 (TT, TE, EE + lowP + lensing + ext)")
))

# Planck 2013 paper XVI Table 5 penultimate column (best fit)
Planck13 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.25886,
    Ob0=0.048252,
    Om0=0.30712,
    H0=67.77 * (u.km / u.s / u.Mpc),
    n=0.9611,
    sigma8=0.8288,
    tau=0.0952,
    z_reion=11.52,
    t0=13.7965 * u.Gyr,
    Tcmb0=2.7255 * u.K,
    Neff=3.046,
    flat=True,
    m_nu=[0., 0., 0.06] * u.eV,
    reference=("Planck Collaboration 2014, A&A, 571, A16 (Paper XVI),"
               " Table 5 (Planck + WP + highL + BAO)")
))


WMAP9 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.2402,
    Ob0=0.04628,
    Om0=0.2865,
    H0=69.32 * (u.km / u.s / u.Mpc),
    n=0.9608,
    sigma8=0.820,
    tau=0.081,
    z_reion=10.1,
    t0=13.772 * u.Gyr,
    Tcmb0=2.725 * u.K,
    Neff=3.04,
    m_nu=0.0 * u.eV,
    flat=True,
    reference=("Hinshaw et al. 2013, ApJS, 208, 19, "
               "doi: 10.1088/0067-0049/208/2/19. "
               "Table 4 (WMAP9 + eCMB + BAO + H0, last column)")
))

WMAP7 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.226,
    Ob0=0.0455,
    Om0=0.272,
    H0=70.4 * (u.km / u.s / u.Mpc),
    n=0.967,
    sigma8=0.810,
    tau=0.085,
    z_reion=10.3,
    t0=13.76 * u.Gyr,
    Tcmb0=2.725 * u.K,
    Neff=3.04,
    m_nu=0.0 * u.eV,
    flat=True,
    reference=("Komatsu et al. 2011, ApJS, 192, 18, "
               "doi: 10.1088/0067-0049/192/2/18. "
               "Table 1 (WMAP + BAO + H0 ML).")
))

WMAP5 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.231,
    Ob0=0.0459,
    Om0=0.277,
    H0=70.2 * (u.km / u.s / u.Mpc),
    n=0.962,
    sigma8=0.817,
    tau=0.088,
    z_reion=11.3,
    t0=13.72 * u.Gyr,
    Tcmb0=2.725 * u.K,
    Neff=3.04,
    m_nu=0.0 * u.eV,
    flat=True,
    reference=("Komatsu et al. 2009, ApJS, 180, 330, "
               "doi: 10.1088/0067-0049/180/2/330. "
               "Table 1 (WMAP + BAO + SN ML).")
))

WMAP3 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.230,
    Ob0=0.0454,
    Om0=0.276,
    H0=70.1 * (u.km / u.s / u.Mpc),
    n=0.946,
    sigma8=0.784,
    tau=0.079,
    z_reion=10.3,
    t0=13.78 * u.Gyr,
    Tcmb0=2.725 * u.K,
    Neff=3.04,
    m_nu=0.0 * u.eV,
    flat=True,
    reference=(r"Spergel et al. 2007, ApJS, 170, 377, "
               r"doi:  10.1086/513700. "
               r"Table 6 (WMAP + SNGold) "
               r"obtained from: https://lambda.gsfc.nasa.gov/product/map/dr2/params/lcdm_wmap_sngold.cfm"
               r"\nPending WMAP team approval and subject to change.")
))

WMAP1 = MappingProxyType(dict(
    cosmology="FlatLambdaCDM",
    Oc0=0.213,
    Ob0=0.0436,
    Om0=0.257,
    H0=72. * (u.km / u.s / u.Mpc),
    n=0.96,
    sigma8=0.75,
    tau=0.117,
    z_reion=17.0,  # Only from WMAP1. Does not exist in the combined analysis.
    t0=13.4 * u.Gyr,
    Tcmb0=2.725 * u.K,
    Neff=3.04,
    m_nu=0.0 * u.eV,
    flat=True,
    reference=(r"Spergel et al. 2003, ApJS, 148, 175, "
               r"doi:  10.1086/377226. "
               r"Table 7 (WMAP + CBI + ACBAR + 2dFGRS + Lya)."
               r"\nPending WMAP team approval and subject to change.")
))


# If new parameters are added, this list must be updated
available = ('Planck18', 'Planck18_arXiv_v2', 'Planck15', 'Planck13',
             'WMAP9', 'WMAP7', 'WMAP5', 'WMAP3', 'WMAP1')



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

# Dependencies
import numpy as np

# Project
from astropy import units as u
from astropy.utils import ShapedLikeNDArray

__all__ = ['Attribute', 'TimeAttribute', 'QuantityAttribute',
           'EarthLocationAttribute', 'CoordinateAttribute',
           'CartesianRepresentationAttribute',
           'DifferentialAttribute']


class Attribute:
    """A non-mutable data descriptor to hold a frame attribute.

    This class must be used to define frame attributes (e.g. ``equinox`` or
    ``obstime``) that are included in a frame class definition.

    Examples
    --------
    The `~astropy.coordinates.FK4` class uses the following class attributes::

      class FK4(BaseCoordinateFrame):
          equinox = TimeAttribute(default=_EQUINOX_B1950)
          obstime = TimeAttribute(default=None,
                                  secondary_attribute='equinox')

    This means that ``equinox`` and ``obstime`` are available to be set as
    keyword arguments when creating an ``FK4`` class instance and are then
    accessible as instance attributes.  The instance value for the attribute
    must be stored in ``'_' + <attribute_name>`` by the frame ``__init__``
    method.

    Note in this example that ``equinox`` and ``obstime`` are time attributes
    and use the ``TimeAttributeFrame`` class.  This subclass overrides the
    ``convert_input`` method to validate and convert inputs into a ``Time``
    object.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    name = '<unbound>'

    def __init__(self, default=None, secondary_attribute=''):
        self.default = default
        self.secondary_attribute = secondary_attribute
        super().__init__()

    def __set_name__(self, owner, name):
        self.name = name

    def convert_input(self, value):
        """
        Validate the input ``value`` and convert to expected attribute class.

        The base method here does nothing, but subclasses can implement this
        as needed.  The method should catch any internal exceptions and raise
        ValueError with an informative message.

        The method returns the validated input along with a boolean that
        indicates whether the input value was actually converted.  If the input
        value was already the correct type then the ``converted`` return value
        should be ``False``.

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        output_value : object
            The ``value`` converted to the correct type (or just ``value`` if
            ``converted`` is False)
        converted : bool
            True if the conversion was actually performed, False otherwise.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.

        """
        return value, False

    def __get__(self, instance, frame_cls=None):
        if instance is None:
            out = self.default
        else:
            out = getattr(instance, '_' + self.name, self.default)
            if out is None:
                out = getattr(instance, self.secondary_attribute, self.default)

        out, converted = self.convert_input(out)
        if instance is not None:
            instance_shape = getattr(instance, 'shape', None)  # None if instance (frame) has no data!
            if instance_shape is not None and (getattr(out, 'shape', ()) and
                                               out.shape != instance_shape):
                # If the shapes do not match, try broadcasting.
                try:
                    if isinstance(out, ShapedLikeNDArray):
                        out = out._apply(np.broadcast_to, shape=instance_shape,
                                         subok=True)
                    else:
                        out = np.broadcast_to(out, instance_shape, subok=True)
                except ValueError:
                    # raise more informative exception.
                    raise ValueError(
                        "attribute {} should be scalar or have shape {}, "
                        "but is has shape {} and could not be broadcast."
                        .format(self.name, instance_shape, out.shape))

                converted = True

            if converted:
                setattr(instance, '_' + self.name, out)

        return out

    def __set__(self, instance, val):
        raise AttributeError('Cannot set frame attribute')


class TimeAttribute(Attribute):
    """
    Frame attribute descriptor for quantities that are Time objects.
    See the `~astropy.coordinates.Attribute` API doc for further
    information.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    def convert_input(self, value):
        """
        Convert input value to a Time object and validate by running through
        the Time constructor.  Also check that the input was a scalar.

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        from astropy.time import Time

        if value is None:
            return None, False

        if isinstance(value, Time):
            out = value
            converted = False
        else:
            try:
                out = Time(value)
            except Exception as err:
                raise ValueError(
                    f'Invalid time input {self.name}={value!r}.') from err
            converted = True

        # Set attribute as read-only for arrays (not allowed by numpy
        # for array scalars)
        if out.shape:
            out.writeable = False
        return out, converted


class CartesianRepresentationAttribute(Attribute):
    """
    A frame attribute that is a CartesianRepresentation with specified units.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    unit : unit-like or None
        Name of a unit that the input will be converted into. If None, no
        unit-checking or conversion is performed
    """

    def __init__(self, default=None, secondary_attribute='', unit=None):
        super().__init__(default, secondary_attribute)
        self.unit = unit

    def convert_input(self, value):
        """
        Checks that the input is a CartesianRepresentation with the correct
        unit, or the special value ``[0, 0, 0]``.

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out : object
            The correctly-typed object.
        converted : boolean
            A boolean which indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        if (isinstance(value, list) and len(value) == 3 and
                all(v == 0 for v in value) and self.unit is not None):
            return CartesianRepresentation(np.zeros(3) * self.unit), True
        else:
            # is it a CartesianRepresentation with correct unit?
            if hasattr(value, 'xyz') and value.xyz.unit == self.unit:
                return value, False

            converted = True
            # if it's a CartesianRepresentation, get the xyz Quantity
            value = getattr(value, 'xyz', value)
            if not hasattr(value, 'unit'):
                raise TypeError('tried to set a {} with something that does '
                                'not have a unit.'
                                .format(self.__class__.__name__))

            value = value.to(self.unit)

            # now try and make a CartesianRepresentation.
            cartrep = CartesianRepresentation(value, copy=False)
            return cartrep, converted


class QuantityAttribute(Attribute):
    """
    A frame attribute that is a quantity with specified units and shape
    (optionally).

    Can be `None`, which should be used for special cases in associated
    frame transformations like "this quantity should be ignored" or similar.

    Parameters
    ----------
    default : number or `~astropy.units.Quantity` or None, optional
        Default value for the attribute if the user does not supply one. If a
        Quantity, it must be consistent with ``unit``, or if a value, ``unit``
        cannot be None.
    secondary_attribute : str, optional
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    unit : unit-like or None, optional
        Name of a unit that the input will be converted into. If None, no
        unit-checking or conversion is performed
    shape : tuple or None, optional
        If given, specifies the shape the attribute must be
    """

    def __init__(self, default=None, secondary_attribute='', unit=None,
                 shape=None):

        if default is None and unit is None:
            raise ValueError('Either a default quantity value must be '
                             'provided, or a unit must be provided to define a '
                             'QuantityAttribute.')

        if default is not None and unit is None:
            unit = default.unit

        self.unit = unit
        self.shape = shape
        default = self.convert_input(default)[0]
        super().__init__(default, secondary_attribute)

    def convert_input(self, value):
        """
        Checks that the input is a Quantity with the necessary units (or the
        special value ``0``).

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        if value is None:
            return None, False

        if (not hasattr(value, 'unit') and self.unit != u.dimensionless_unscaled
                and np.any(value != 0)):
            raise TypeError('Tried to set a QuantityAttribute with '
                            'something that does not have a unit.')

        oldvalue = value
        value = u.Quantity(oldvalue, self.unit, copy=False)
        if self.shape is not None and value.shape != self.shape:
            if value.shape == () and oldvalue == 0:
                # Allow a single 0 to fill whatever shape is needed.
                value = np.broadcast_to(value, self.shape, subok=True)
            else:
                raise ValueError(
                    f'The provided value has shape "{value.shape}", but '
                    f'should have shape "{self.shape}"')

        converted = oldvalue is not value
        return value, converted


class EarthLocationAttribute(Attribute):
    """
    A frame attribute that can act as a `~astropy.coordinates.EarthLocation`.
    It can be created as anything that can be transformed to the
    `~astropy.coordinates.ITRS` frame, but always presents as an `EarthLocation`
    when accessed after creation.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    def convert_input(self, value):
        """
        Checks that the input is a Quantity with the necessary units (or the
        special value ``0``).

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        if value is None:
            return None, False
        elif isinstance(value, EarthLocation):
            return value, False
        else:
            # we have to do the import here because of some tricky circular deps
            from .builtin_frames import ITRS

            if not hasattr(value, 'transform_to'):
                raise ValueError('"{}" was passed into an '
                                 'EarthLocationAttribute, but it does not have '
                                 '"transform_to" method'.format(value))
            itrsobj = value.transform_to(ITRS())
            return itrsobj.earth_location, True


class CoordinateAttribute(Attribute):
    """
    A frame attribute which is a coordinate object.  It can be given as a
    `~astropy.coordinates.SkyCoord` or a low-level frame instance.  If a
    low-level frame instance is provided, it will always be upgraded to be a
    `~astropy.coordinates.SkyCoord` to ensure consistent transformation
    behavior.  The coordinate object will always be returned as a low-level
    frame instance when accessed.

    Parameters
    ----------
    frame : `~astropy.coordinates.BaseCoordinateFrame` class
        The type of frame this attribute can be
    default : object
        Default value for the attribute if not provided
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    def __init__(self, frame, default=None, secondary_attribute=''):
        self._frame = frame
        super().__init__(default, secondary_attribute)

    def convert_input(self, value):
        """
        Checks that the input is a SkyCoord with the necessary units (or the
        special value ``None``).

        Parameters
        ----------
        value : object
            Input value to be converted.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """
        from astropy.coordinates import SkyCoord

        if value is None:
            return None, False
        elif isinstance(value, self._frame):
            return value, False
        else:
            value = SkyCoord(value)  # always make the value a SkyCoord
            transformedobj = value.transform_to(self._frame)
            return transformedobj.frame, True


class DifferentialAttribute(Attribute):
    """A frame attribute which is a differential instance.

    The optional ``allowed_classes`` argument allows specifying a restricted
    set of valid differential classes to check the input against. Otherwise,
    any `~astropy.coordinates.BaseDifferential` subclass instance is valid.

    Parameters
    ----------
    default : object
        Default value for the attribute if not provided
    allowed_classes : tuple, optional
        A list of allowed differential classes for this attribute to have.
    secondary_attribute : str
        Name of a secondary instance attribute which supplies the value if
        ``default is None`` and no value was supplied during initialization.
    """

    def __init__(self, default=None, allowed_classes=None,
                 secondary_attribute=''):

        if allowed_classes is not None:
            self.allowed_classes = tuple(allowed_classes)
        else:
            self.allowed_classes = BaseDifferential

        super().__init__(default, secondary_attribute)

    def convert_input(self, value):
        """
        Checks that the input is a differential object and is one of the
        allowed class types.

        Parameters
        ----------
        value : object
            Input value.

        Returns
        -------
        out, converted : correctly-typed object, boolean
            Tuple consisting of the correctly-typed object and a boolean which
            indicates if conversion was actually performed.

        Raises
        ------
        ValueError
            If the input is not valid for this attribute.
        """

        if value is None:
            return None, False

        if not isinstance(value, self.allowed_classes):
            if len(self.allowed_classes) == 1:
                value = self.allowed_classes[0](value)
            else:
                raise TypeError('Tried to set a DifferentialAttribute with '
                                'an unsupported Differential type {}. Allowed '
                                'classes are: {}'
                                .format(value.__class__,
                                        self.allowed_classes))

        return value, True


# do this here to prevent a series of complicated circular imports
from .earth import EarthLocation
from .representation import CartesianRepresentation, BaseDifferential



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
import pickle

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal

from astropy.utils.data import get_pkg_data_contents, get_pkg_data_fileobj
from astropy.utils.exceptions import AstropyDeprecationWarning
from astropy.utils.misc import NumpyRNGContext
from astropy.io import fits
from astropy.io.fits.verify import VerifyWarning
from astropy import wcs
from astropy.wcs.wcs import FITSFixedWarning


def test_basic():
    wcs1 = wcs.WCS()
    s = pickle.dumps(wcs1)
    with pytest.warns(FITSFixedWarning):
        pickle.loads(s)


def test_dist():
    with get_pkg_data_fileobj(
            os.path.join("data", "dist.fits"), encoding='binary') as test_file:
        hdulist = fits.open(test_file)
        # The use of ``AXISCORR`` for D2IM correction has been deprecated
        with pytest.warns(AstropyDeprecationWarning):
            wcs1 = wcs.WCS(hdulist[0].header, hdulist)
        assert wcs1.det2im2 is not None

        s = pickle.dumps(wcs1)
        with pytest.warns(FITSFixedWarning):
            wcs2 = pickle.loads(s)

        with NumpyRNGContext(123456789):
            x = np.random.rand(2 ** 16, wcs1.wcs.naxis)
            world1 = wcs1.all_pix2world(x, 1)
            world2 = wcs2.all_pix2world(x, 1)

        assert_array_almost_equal(world1, world2)


def test_sip():
    with get_pkg_data_fileobj(
            os.path.join("data", "sip.fits"), encoding='binary') as test_file:
        hdulist = fits.open(test_file, ignore_missing_end=True)
        with pytest.warns(FITSFixedWarning):
            wcs1 = wcs.WCS(hdulist[0].header)
        assert wcs1.sip is not None
        s = pickle.dumps(wcs1)
        with pytest.warns(FITSFixedWarning):
            wcs2 = pickle.loads(s)

        with NumpyRNGContext(123456789):
            x = np.random.rand(2 ** 16, wcs1.wcs.naxis)
            world1 = wcs1.all_pix2world(x, 1)
            world2 = wcs2.all_pix2world(x, 1)

        assert_array_almost_equal(world1, world2)


def test_sip2():
    with get_pkg_data_fileobj(
            os.path.join("data", "sip2.fits"), encoding='binary') as test_file:
        hdulist = fits.open(test_file, ignore_missing_end=True)
        with pytest.warns(FITSFixedWarning):
            wcs1 = wcs.WCS(hdulist[0].header)
        assert wcs1.sip is not None
        s = pickle.dumps(wcs1)
        with pytest.warns(FITSFixedWarning):
            wcs2 = pickle.loads(s)

        with NumpyRNGContext(123456789):
            x = np.random.rand(2 ** 16, wcs1.wcs.naxis)
            world1 = wcs1.all_pix2world(x, 1)
            world2 = wcs2.all_pix2world(x, 1)

        assert_array_almost_equal(world1, world2)


# Ignore "PV2_2 = 0.209028857410973 invalid keyvalue" warning seen on Windows.
@pytest.mark.filterwarnings(r'ignore:PV2_2')
def test_wcs():
    header = get_pkg_data_contents(
        os.path.join("data", "outside_sky.hdr"), encoding='binary')

    wcs1 = wcs.WCS(header)
    s = pickle.dumps(wcs1)
    with pytest.warns(FITSFixedWarning):
        wcs2 = pickle.loads(s)

    with NumpyRNGContext(123456789):
        x = np.random.rand(2 ** 16, wcs1.wcs.naxis)
        world1 = wcs1.all_pix2world(x, 1)
        world2 = wcs2.all_pix2world(x, 1)

    assert_array_almost_equal(world1, world2)


class Sub(wcs.WCS):
    def __init__(self, *args, **kwargs):
        self.foo = 42


def test_subclass():
    wcs = Sub()
    s = pickle.dumps(wcs)
    with pytest.warns(FITSFixedWarning):
        wcs2 = pickle.loads(s)

    assert isinstance(wcs2, Sub)
    assert wcs.foo == 42
    assert wcs2.foo == 42
    assert wcs2.wcs is not None



# -*- coding: utf-8 -*-

"""
Readers, Writers, and I/O Miscellany.
"""

__all__ = ["file_reader", "file_writer"]  # from `mypackage`

# e.g. a file reader and writer for ``myformat``
# this will be used in ``astropy_io.py``
from .core import file_reader, file_writer

# Register read and write methods into Astropy:
# determine if it is 1) installed and 2) the correct version (v5.0+)
try:
    import astropy
    from astropy.utils.introspection import minversion
except ImportError:
    ASTROPY_GE_5 = False
else:
    ASTROPY_GE_5 = minversion(astropy, "5.0")

if ASTROPY_GE_5:
    # Astropy is installed and v5.0+ so we import the following modules
    # to register "myformat" with Cosmology read/write and "mypackage"
    # with Cosmology to/from_format.
    from . import astropy_convert, astropy_io



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from numpy.testing import assert_allclose

from astropy.utils.compat.optional_deps import HAS_PLT, HAS_SCIPY
if HAS_PLT:
    import matplotlib.pyplot as plt

import pytest
import numpy as np

from astropy.visualization import hist
from astropy.stats import histogram


@pytest.mark.skipif('not HAS_PLT')
def test_hist_basic(rseed=0):
    rng = np.random.default_rng(rseed)
    x = rng.standard_normal(100)

    for range in [None, (-2, 2)]:
        n1, bins1, patches1 = plt.hist(x, 10, range=range)
        n2, bins2, patches2 = hist(x, 10, range=range)

        assert_allclose(n1, n2)
        assert_allclose(bins1, bins2)


@pytest.mark.skipif('not HAS_PLT')
def test_hist_specify_ax(rseed=0):
    rng = np.random.default_rng(rseed)
    x = rng.standard_normal(100)

    fig, ax = plt.subplots(2)
    n1, bins1, patches1 = hist(x, 10, ax=ax[0])
    assert patches1[0].axes is ax[0]

    n2, bins2, patches2 = hist(x, 10, ax=ax[1])
    assert patches2[0].axes is ax[1]


@pytest.mark.skipif('not HAS_PLT')
def test_hist_autobin(rseed=0):
    rng = np.random.default_rng(rseed)
    x = rng.standard_normal(100)

    # 'knuth' bintype depends on scipy that is optional dependency
    if HAS_SCIPY:
        bintypes = [10, np.arange(-3, 3, 10), 'knuth', 'scott',
                    'freedman', 'blocks']
    else:
        bintypes = [10, np.arange(-3, 3, 10), 'scott',
                    'freedman', 'blocks']

    for bintype in bintypes:
        for range in [None, (-3, 3)]:
            n1, bins1 = histogram(x, bintype, range=range)
            n2, bins2, patches = hist(x, bintype, range=range)
            assert_allclose(n1, n2)
            assert_allclose(bins1, bins2)


def test_histogram_pathological_input():
    # Regression test for https://github.com/astropy/astropy/issues/7758

    # The key feature of the data below is that one of the points is very,
    # very different than the rest. That leads to a large number of bins.
    data = [9.99999914e+05, -8.31312483e-03, 6.52755852e-02, 1.43104653e-03,
            -2.26311017e-02, 2.82660007e-03, 1.80307521e-02, 9.26294279e-03,
            5.06606026e-02, 2.05418011e-03]

    with pytest.raises(ValueError):
        hist(data, bins='freedman', max_bins=10000)



# Licensed under a 3-clause BSD style license - see LICENSE.rst


from .core import *  # noqa
from .kernels import *  # noqa
from .utils import discretize_model, KernelSizeError  # noqa

from .convolve import convolve, convolve_fft, interpolate_replace_nans, convolve_models, convolve_models_fft  # noqa

# Deprecated kernels that are not defined in __all__
from .kernels import MexicanHat1DKernel, MexicanHat2DKernel



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from numpy.testing import assert_array_equal

from astropy.modeling import powerlaws
from .basic import TransformType
from . import _parameter_to_value


__all__ = ['PowerLaw1DType', 'BrokenPowerLaw1DType',
           'SmoothlyBrokenPowerLaw1DType', 'ExponentialCutoffPowerLaw1DType',
           'LogParabola1DType']


class PowerLaw1DType(TransformType):
    name = 'transform/power_law1d'
    version = '1.0.0'
    types = ['astropy.modeling.powerlaws.PowerLaw1D']

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return powerlaws.PowerLaw1D(amplitude=node['amplitude'],
                                    x_0=node['x_0'],
                                    alpha=node['alpha'])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {'amplitude': _parameter_to_value(model.amplitude),
                'x_0': _parameter_to_value(model.x_0),
                'alpha': _parameter_to_value(model.alpha)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert (isinstance(a, powerlaws.PowerLaw1D) and
                isinstance(b, powerlaws.PowerLaw1D))
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.alpha, b.alpha)


class BrokenPowerLaw1DType(TransformType):
    name = 'transform/broken_power_law1d'
    version = '1.0.0'
    types = ['astropy.modeling.powerlaws.BrokenPowerLaw1D']

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return powerlaws.BrokenPowerLaw1D(amplitude=node['amplitude'],
                                          x_break=node['x_break'],
                                          alpha_1=node['alpha_1'],
                                          alpha_2=node['alpha_2'])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {'amplitude': _parameter_to_value(model.amplitude),
                'x_break': _parameter_to_value(model.x_break),
                'alpha_1': _parameter_to_value(model.alpha_1),
                'alpha_2': _parameter_to_value(model.alpha_2)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert (isinstance(a, powerlaws.BrokenPowerLaw1D) and
                isinstance(b, powerlaws.BrokenPowerLaw1D))
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_break, b.x_break)
        assert_array_equal(a.alpha_1, b.alpha_1)
        assert_array_equal(a.alpha_2, b.alpha_2)


class SmoothlyBrokenPowerLaw1DType(TransformType):
    name = 'transform/smoothly_broken_power_law1d'
    version = '1.0.0'
    types = ['astropy.modeling.powerlaws.SmoothlyBrokenPowerLaw1D']

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return powerlaws.SmoothlyBrokenPowerLaw1D(amplitude=node['amplitude'],
                                                  x_break=node['x_break'],
                                                  alpha_1=node['alpha_1'],
                                                  alpha_2=node['alpha_2'],
                                                  delta=node['delta'])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {'amplitude': _parameter_to_value(model.amplitude),
                'x_break': _parameter_to_value(model.x_break),
                'alpha_1': _parameter_to_value(model.alpha_1),
                'alpha_2': _parameter_to_value(model.alpha_2),
                'delta': _parameter_to_value(model.delta)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert (isinstance(a, powerlaws.SmoothlyBrokenPowerLaw1D) and
                isinstance(b, powerlaws.SmoothlyBrokenPowerLaw1D))
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_break, b.x_break)
        assert_array_equal(a.alpha_1, b.alpha_1)
        assert_array_equal(a.alpha_2, b.alpha_2)
        assert_array_equal(a.delta, b.delta)


class ExponentialCutoffPowerLaw1DType(TransformType):
    name = 'transform/exponential_cutoff_power_law1d'
    version = '1.0.0'
    types = ['astropy.modeling.powerlaws.ExponentialCutoffPowerLaw1D']

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return powerlaws.ExponentialCutoffPowerLaw1D(amplitude=node['amplitude'],
                                                     x_0=node['x_0'],
                                                     alpha=node['alpha'],
                                                     x_cutoff=node['x_cutoff'])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {'amplitude': _parameter_to_value(model.amplitude),
                'x_0': _parameter_to_value(model.x_0),
                'alpha': _parameter_to_value(model.alpha),
                'x_cutoff': _parameter_to_value(model.x_cutoff)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert (isinstance(a, powerlaws.ExponentialCutoffPowerLaw1D) and
                isinstance(b, powerlaws.ExponentialCutoffPowerLaw1D))
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.alpha, b.alpha)
        assert_array_equal(a.x_cutoff, b.x_cutoff)


class LogParabola1DType(TransformType):
    name = 'transform/log_parabola1d'
    version = '1.0.0'
    types = ['astropy.modeling.powerlaws.LogParabola1D']

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return powerlaws.LogParabola1D(amplitude=node['amplitude'],
                                       x_0=node['x_0'],
                                       alpha=node['alpha'],
                                       beta=node['beta'])

    @classmethod
    def to_tree_transform(cls, model, ctx):
        node = {'amplitude': _parameter_to_value(model.amplitude),
                'x_0': _parameter_to_value(model.x_0),
                'alpha': _parameter_to_value(model.alpha),
                'beta': _parameter_to_value(model.beta)}
        return node

    @classmethod
    def assert_equal(cls, a, b):
        # TODO: If models become comparable themselves, remove this.
        TransformType.assert_equal(a, b)
        assert (isinstance(a, powerlaws.LogParabola1D) and
                isinstance(b, powerlaws.LogParabola1D))
        assert_array_equal(a.amplitude, b.amplitude)
        assert_array_equal(a.x_0, b.x_0)
        assert_array_equal(a.alpha, b.alpha)
        assert_array_equal(a.beta, b.beta)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""
This module implements classes (called Fitters) which combine optimization
algorithms (typically from `scipy.optimize`) with statistic functions to perform
fitting. Fitters are implemented as callable classes. In addition to the data
to fit, the ``__call__`` method takes an instance of
`~astropy.modeling.core.FittableModel` as input, and returns a copy of the
model with its parameters determined by the optimizer.

Optimization algorithms, called "optimizers" are implemented in
`~astropy.modeling.optimizers` and statistic functions are in
`~astropy.modeling.statistic`. The goal is to provide an easy to extend
framework and allow users to easily create new fitters by combining statistics
with optimizers.

There are two exceptions to the above scheme.
`~astropy.modeling.fitting.LinearLSQFitter` uses Numpy's `~numpy.linalg.lstsq`
function.  `~astropy.modeling.fitting.LevMarLSQFitter` uses
`~scipy.optimize.leastsq` which combines optimization and statistic in one
implementation.
"""
# pylint: disable=invalid-name

import abc
import inspect
import operator
import warnings
from importlib.metadata import entry_points

from functools import reduce, wraps

import numpy as np

from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
from .utils import poly_map_domain, _combine_equivalency_dict
from .optimizers import (SLSQP, Simplex)
from .statistic import (leastsquare)
from .optimizers import (DEFAULT_MAXITER, DEFAULT_EPS, DEFAULT_ACC)
from .spline import (SplineInterpolateFitter, SplineSmoothingFitter,
                     SplineExactKnotsFitter, SplineSplrepFitter)

__all__ = ['LinearLSQFitter', 'LevMarLSQFitter', 'FittingWithOutlierRemoval',
           'SLSQPLSQFitter', 'SimplexLSQFitter', 'JointFitter', 'Fitter',
           "ModelLinearityError", "ModelsError"]


# Statistic functions implemented in `astropy.modeling.statistic.py
STATISTICS = [leastsquare]

# Optimizers implemented in `astropy.modeling.optimizers.py
OPTIMIZERS = [Simplex, SLSQP]


class Covariance():
    """Class for covariance matrix calculated by fitter. """

    def __init__(self, cov_matrix, param_names):
        self.cov_matrix = cov_matrix
        self.param_names = param_names

    def pprint(self, max_lines, round_val):
        # Print and label lower triangle of covariance matrix
        # Print rows for params up to `max_lines`, round floats to 'round_val'
        longest_name = max([len(x) for x in self.param_names])
        ret_str = 'parameter variances / covariances \n'
        fstring = f'{"": <{longest_name}}| {{0}}\n'
        for i, row in enumerate(self.cov_matrix):
            if i <= max_lines-1:
                param = self.param_names[i]
                ret_str += fstring.replace(' '*len(param), param, 1).\
                           format(repr(np.round(row[:i+1], round_val))[7:-2])
            else:
                ret_str += '...'
        return(ret_str.rstrip())

    def __repr__(self):
        return(self.pprint(max_lines=10, round_val=3))

    def __getitem__(self, params):
        # index covariance matrix by parameter names or indices
        if len(params) != 2:
            raise ValueError('Covariance must be indexed by two values.')
        if all(isinstance(item, str) for item in params):
            i1, i2 = self.param_names.index(params[0]), self.param_names.index(params[1])
        elif all(isinstance(item, int) for item in params):
            i1, i2 = params
        else:
            raise TypeError('Covariance can be indexed by two parameter names or integer indices.')
        return(self.cov_matrix[i1][i2])


class StandardDeviations():
    """ Class for fitting uncertainties."""

    def __init__(self, cov_matrix, param_names):
        self.param_names = param_names
        self.stds = self._calc_stds(cov_matrix)

    def _calc_stds(self, cov_matrix):
        # sometimes scipy lstsq returns a non-sensical negative vals in the
        # diagonals of the cov_x it computes.
        stds = [np.sqrt(x) if x > 0 else None for x in np.diag(cov_matrix)]
        return stds

    def pprint(self, max_lines, round_val):
        longest_name = max([len(x) for x in self.param_names])
        ret_str = 'standard deviations\n'
        fstring = '{0}{1}| {2}\n'
        for i, std in enumerate(self.stds):
            if i <= max_lines-1:
                param = self.param_names[i]
                ret_str += fstring.format(param,
                                          ' ' * (longest_name - len(param)),
                                          str(np.round(std, round_val)))
            else:
                ret_str += '...'
        return(ret_str.rstrip())

    def __repr__(self):
        return(self.pprint(max_lines=10, round_val=3))

    def __getitem__(self, param):
        if isinstance(param, str):
            i = self.param_names.index(param)
        elif isinstance(param, int):
            i = param
        else:
            raise TypeError('Standard deviation can be indexed by parameter name or integer.')
        return(self.stds[i])


class ModelsError(Exception):
    """Base class for model exceptions"""


class ModelLinearityError(ModelsError):
    """ Raised when a non-linear model is passed to a linear fitter."""


class UnsupportedConstraintError(ModelsError, ValueError):
    """
    Raised when a fitter does not support a type of constraint.
    """


class _FitterMeta(abc.ABCMeta):
    """
    Currently just provides a registry for all Fitter classes.
    """

    registry = set()

    def __new__(mcls, name, bases, members):
        cls = super().__new__(mcls, name, bases, members)

        if not inspect.isabstract(cls) and not name.startswith('_'):
            mcls.registry.add(cls)

        return cls


def fitter_unit_support(func):
    """
    This is a decorator that can be used to add support for dealing with
    quantities to any __call__ method on a fitter which may not support
    quantities itself. This is done by temporarily removing units from all
    parameters then adding them back once the fitting has completed.
    """
    @wraps(func)
    def wrapper(self, model, x, y, z=None, **kwargs):
        equivalencies = kwargs.pop('equivalencies', None)

        data_has_units = (isinstance(x, Quantity) or
                          isinstance(y, Quantity) or
                          isinstance(z, Quantity))

        model_has_units = model._has_units

        if data_has_units or model_has_units:

            if model._supports_unit_fitting:

                # We now combine any instance-level input equivalencies with user
                # specified ones at call-time.

                input_units_equivalencies = _combine_equivalency_dict(
                    model.inputs, equivalencies, model.input_units_equivalencies)

                # If input_units is defined, we transform the input data into those
                # expected by the model. We hard-code the input names 'x', and 'y'
                # here since FittableModel instances have input names ('x',) or
                # ('x', 'y')

                if model.input_units is not None:
                    if isinstance(x, Quantity):
                        x = x.to(model.input_units[model.inputs[0]],
                                 equivalencies=input_units_equivalencies[model.inputs[0]])
                    if isinstance(y, Quantity) and z is not None:
                        y = y.to(model.input_units[model.inputs[1]],
                                 equivalencies=input_units_equivalencies[model.inputs[1]])

                # Create a dictionary mapping the real model inputs and outputs
                # names to the data. This remapping of names must be done here, after
                # the input data is converted to the correct units.
                rename_data = {model.inputs[0]: x}
                if z is not None:
                    rename_data[model.outputs[0]] = z
                    rename_data[model.inputs[1]] = y
                else:
                    rename_data[model.outputs[0]] = y
                    rename_data['z'] = None

                # We now strip away the units from the parameters, taking care to
                # first convert any parameters to the units that correspond to the
                # input units (to make sure that initial guesses on the parameters)
                # are in the right unit system
                model = model.without_units_for_data(**rename_data)
                # We strip away the units from the input itself
                add_back_units = False

                if isinstance(x, Quantity):
                    add_back_units = True
                    xdata = x.value
                else:
                    xdata = np.asarray(x)

                if isinstance(y, Quantity):
                    add_back_units = True
                    ydata = y.value
                else:
                    ydata = np.asarray(y)

                if z is not None:
                    if isinstance(z, Quantity):
                        add_back_units = True
                        zdata = z.value
                    else:
                        zdata = np.asarray(z)
                # We run the fitting
                if z is None:
                    model_new = func(self, model, xdata, ydata, **kwargs)
                else:
                    model_new = func(self, model, xdata, ydata, zdata, **kwargs)

                # And finally we add back units to the parameters
                if add_back_units:
                    model_new = model_new.with_units_from_data(**rename_data)
                return model_new

            else:

                raise NotImplementedError("This model does not support being "
                                          "fit to data with units.")

        else:

            return func(self, model, x, y, z=z, **kwargs)

    return wrapper


class Fitter(metaclass=_FitterMeta):
    """
    Base class for all fitters.

    Parameters
    ----------
    optimizer : callable
        A callable implementing an optimization algorithm
    statistic : callable
        Statistic function

    """

    supported_constraints = []

    def __init__(self, optimizer, statistic):
        if optimizer is None:
            raise ValueError("Expected an optimizer.")
        if statistic is None:
            raise ValueError("Expected a statistic function.")
        if inspect.isclass(optimizer):
            # a callable class
            self._opt_method = optimizer()
        elif inspect.isfunction(optimizer):
            self._opt_method = optimizer
        else:
            raise ValueError("Expected optimizer to be a callable class or a function.")
        if inspect.isclass(statistic):
            self._stat_method = statistic()
        else:
            self._stat_method = statistic

    def objective_function(self, fps, *args):
        """
        Function to minimize.

        Parameters
        ----------
        fps : list
            parameters returned by the fitter
        args : list
            [model, [other_args], [input coordinates]]
            other_args may include weights or any other quantities specific for
            a statistic

        Notes
        -----
        The list of arguments (args) is set in the `__call__` method.
        Fitters may overwrite this method, e.g. when statistic functions
        require other arguments.

        """
        model = args[0]
        meas = args[-1]
        _fitter_to_model_params(model, fps)
        res = self._stat_method(meas, model, *args[1:-1])
        return res

    @staticmethod
    def _add_fitting_uncertainties(*args):
        """
        When available, calculate and sets the parameter covariance matrix
        (model.cov_matrix) and standard deviations (model.stds).
        """
        return None

    @abc.abstractmethod
    def __call__(self):
        """
        This method performs the actual fitting and modifies the parameter list
        of a model.
        Fitter subclasses should implement this method.
        """

        raise NotImplementedError("Subclasses should implement this method.")


# TODO: I have ongoing branch elsewhere that's refactoring this module so that
# all the fitter classes in here are Fitter subclasses.  In the meantime we
# need to specify that _FitterMeta is its metaclass.
class LinearLSQFitter(metaclass=_FitterMeta):
    """
    A class performing a linear least square fitting.
    Uses `numpy.linalg.lstsq` to do the fitting.
    Given a model and data, fits the model to the data and changes the
    model's parameters. Keeps a dictionary of auxiliary fitting information.
    Notes
    -----
    Note that currently LinearLSQFitter does not support compound models.
    """

    supported_constraints = ['fixed']
    supports_masked_input = True

    def __init__(self, calc_uncertainties=False):
        self.fit_info = {'residuals': None,
                         'rank': None,
                         'singular_values': None,
                         'params': None
                         }
        self._calc_uncertainties=calc_uncertainties

    @staticmethod
    def _is_invertible(m):
        """Check if inverse of matrix can be obtained."""
        if m.shape[0] != m.shape[1]:
            return False
        if np.linalg.matrix_rank(m) < m.shape[0]:
            return False
        return True

    def _add_fitting_uncertainties(self, model, a, n_coeff, x, y, z=None,
                                   resids=None):
        """
        Calculate and parameter covariance matrix and standard deviations
        and set `cov_matrix` and `stds` attributes.
        """
        x_dot_x_prime = np.dot(a.T, a)
        masked = False or hasattr(y, 'mask')

        # check if invertible. if not, can't calc covariance.
        if not self._is_invertible(x_dot_x_prime):
            return(model)
        inv_x_dot_x_prime = np.linalg.inv(x_dot_x_prime)

        if z is None:  # 1D models
            if len(model) == 1:  # single model
                mask = None
                if masked:
                    mask = y.mask
                xx = np.ma.array(x, mask=mask)
                RSS = [(1/(xx.count()-n_coeff)) * resids]

            if len(model) > 1:  # model sets
                RSS = []   # collect sum residuals squared for each model in set
                for j in range(len(model)):
                    mask = None
                    if masked:
                        mask = y.mask[..., j].flatten()
                    xx = np.ma.array(x, mask=mask)
                    eval_y = model(xx, model_set_axis=False)
                    eval_y = np.rollaxis(eval_y, model.model_set_axis)[j]
                    RSS.append((1/(xx.count()-n_coeff)) * np.sum((y[..., j] - eval_y)**2))

        else:  # 2D model
            if len(model) == 1:
                mask = None
                if masked:
                    warnings.warn('Calculation of fitting uncertainties '
                                  'for 2D models with masked values not '
                                  'currently supported.\n',
                                  AstropyUserWarning)
                    return
                xx, yy = np.ma.array(x, mask=mask), np.ma.array(y, mask=mask)
                # len(xx) instead of xx.count. this will break if values are masked?
                RSS = [(1/(len(xx)-n_coeff)) * resids]
            else:
                RSS = []
                for j in range(len(model)):
                    eval_z = model(x, y, model_set_axis=False)
                    mask = None  # need to figure out how to deal w/ masking here.
                    if model.model_set_axis == 1:
                        # model_set_axis passed when evaluating only refers to input shapes
                        # so output must be reshaped for model_set_axis=1.
                        eval_z = np.rollaxis(eval_z, 1)
                    eval_z = eval_z[j]
                    RSS.append([(1/(len(x)-n_coeff)) * np.sum((z[j] - eval_z)**2)])

        covs = [inv_x_dot_x_prime * r for r in RSS]
        free_param_names = [x for x in model.fixed if (model.fixed[x] is False)
                            and (model.tied[x] is False)]

        if len(covs) == 1:
            model.cov_matrix = Covariance(covs[0], model.param_names)
            model.stds = StandardDeviations(covs[0], free_param_names)
        else:
            model.cov_matrix = [Covariance(cov, model.param_names) for cov in covs]
            model.stds = [StandardDeviations(cov, free_param_names) for cov in covs]

    @staticmethod
    def _deriv_with_constraints(model, param_indices, x=None, y=None):
        if y is None:
            d = np.array(model.fit_deriv(x, *model.parameters))
        else:
            d = np.array(model.fit_deriv(x, y, *model.parameters))

        if model.col_fit_deriv:
            return d[param_indices]
        else:
            return d[..., param_indices]

    def _map_domain_window(self, model, x, y=None):
        """
        Maps domain into window for a polynomial model which has these
        attributes.
        """

        if y is None:
            if hasattr(model, 'domain') and model.domain is None:
                model.domain = [x.min(), x.max()]
            if hasattr(model, 'window') and model.window is None:
                model.window = [-1, 1]
            return poly_map_domain(x, model.domain, model.window)
        else:
            if hasattr(model, 'x_domain') and model.x_domain is None:
                model.x_domain = [x.min(), x.max()]
            if hasattr(model, 'y_domain') and model.y_domain is None:
                model.y_domain = [y.min(), y.max()]
            if hasattr(model, 'x_window') and model.x_window is None:
                model.x_window = [-1., 1.]
            if hasattr(model, 'y_window') and model.y_window is None:
                model.y_window = [-1., 1.]

            xnew = poly_map_domain(x, model.x_domain, model.x_window)
            ynew = poly_map_domain(y, model.y_domain, model.y_window)
            return xnew, ynew

    @fitter_unit_support
    def __call__(self, model, x, y, z=None, weights=None, rcond=None):
        """
        Fit data to this model.

        Parameters
        ----------
        model : `~astropy.modeling.FittableModel`
            model to fit to x, y, z
        x : array
            Input coordinates
        y : array-like
            Input coordinates
        z : array-like, optional
            Input coordinates.
            If the dependent (``y`` or ``z``) coordinate values are provided
            as a `numpy.ma.MaskedArray`, any masked points are ignored when
            fitting. Note that model set fitting is significantly slower when
            there are masked points (not just an empty mask), as the matrix
            equation has to be solved for each model separately when their
            coordinate grids differ.
        weights : array, optional
            Weights for fitting.
            For data with Gaussian uncertainties, the weights should be
            1/sigma.
        rcond :  float, optional
            Cut-off ratio for small singular values of ``a``.
            Singular values are set to zero if they are smaller than ``rcond``
            times the largest singular value of ``a``.
        equivalencies : list or None, optional, keyword-only
            List of *additional* equivalencies that are should be applied in
            case x, y and/or z have units. Default is None.

        Returns
        -------
        model_copy : `~astropy.modeling.FittableModel`
            a copy of the input model with parameters set by the fitter

        """

        if not model.fittable:
            raise ValueError("Model must be a subclass of FittableModel")

        if not model.linear:
            raise ModelLinearityError('Model is not linear in parameters, '
                                      'linear fit methods should not be used.')

        if hasattr(model, "submodel_names"):
            raise ValueError("Model must be simple, not compound")

        _validate_constraints(self.supported_constraints, model)

        model_copy = model.copy()
        model_copy.sync_constraints = False
        _, fitparam_indices = _model_to_fit_params(model_copy)

        if model_copy.n_inputs == 2 and z is None:
            raise ValueError("Expected x, y and z for a 2 dimensional model.")

        farg = _convert_input(x, y, z, n_models=len(model_copy),
                              model_set_axis=model_copy.model_set_axis)

        has_fixed = any(model_copy.fixed.values())

        # This is also done by _convert_inputs, but we need it here to allow
        # checking the array dimensionality before that gets called:
        if weights is not None:
            weights = np.asarray(weights, dtype=float)

        if has_fixed:

            # The list of fixed params is the complement of those being fitted:
            fixparam_indices = [idx for idx in
                                range(len(model_copy.param_names))
                                if idx not in fitparam_indices]

            # Construct matrix of user-fixed parameters that can be dotted with
            # the corresponding fit_deriv() terms, to evaluate corrections to
            # the dependent variable in order to fit only the remaining terms:
            fixparams = np.asarray([getattr(model_copy,
                                            model_copy.param_names[idx]).value
                                    for idx in fixparam_indices])

        if len(farg) == 2:
            x, y = farg

            if weights is not None:
                # If we have separate weights for each model, apply the same
                # conversion as for the data, otherwise check common weights
                # as if for a single model:
                _, weights = _convert_input(
                    x, weights,
                    n_models=len(model_copy) if weights.ndim == y.ndim else 1,
                    model_set_axis=model_copy.model_set_axis
                )

            # map domain into window
            if hasattr(model_copy, 'domain'):
                x = self._map_domain_window(model_copy, x)
            if has_fixed:
                lhs = np.asarray(self._deriv_with_constraints(model_copy,
                                                              fitparam_indices,
                                                              x=x))
                fixderivs = self._deriv_with_constraints(model_copy, fixparam_indices, x=x)
            else:
                lhs = np.asarray(model_copy.fit_deriv(x, *model_copy.parameters))
            sum_of_implicit_terms = model_copy.sum_of_implicit_terms(x)
            rhs = y
        else:
            x, y, z = farg

            if weights is not None:
                # If we have separate weights for each model, apply the same
                # conversion as for the data, otherwise check common weights
                # as if for a single model:
                _, _, weights = _convert_input(
                    x, y, weights,
                    n_models=len(model_copy) if weights.ndim == z.ndim else 1,
                    model_set_axis=model_copy.model_set_axis
                )

            # map domain into window
            if hasattr(model_copy, 'x_domain'):
                x, y = self._map_domain_window(model_copy, x, y)

            if has_fixed:
                lhs = np.asarray(self._deriv_with_constraints(model_copy,
                                                              fitparam_indices, x=x, y=y))
                fixderivs = self._deriv_with_constraints(model_copy,
                                                         fixparam_indices,
                                                         x=x, y=y)
            else:
                lhs = np.asanyarray(model_copy.fit_deriv(x, y, *model_copy.parameters))
            sum_of_implicit_terms = model_copy.sum_of_implicit_terms(x, y)

            if len(model_copy) > 1:

                # Just to be explicit (rather than baking in False == 0):
                model_axis = model_copy.model_set_axis or 0

                if z.ndim > 2:
                    # For higher-dimensional z, flatten all the axes except the
                    # dimension along which models are stacked and transpose so
                    # the model axis is *last* (I think this resolves Erik's
                    # pending generalization from 80a6f25a):
                    rhs = np.rollaxis(z, model_axis, z.ndim)
                    rhs = rhs.reshape(-1, rhs.shape[-1])
                else:
                    # This "else" seems to handle the corner case where the
                    # user has already flattened x/y before attempting a 2D fit
                    # but z has a second axis for the model set. NB. This is
                    # ~5-10x faster than using rollaxis.
                    rhs = z.T if model_axis == 0 else z

                if weights is not None:
                    # Same for weights
                    if weights.ndim > 2:
                        # Separate 2D weights for each model:
                        weights = np.rollaxis(weights, model_axis, weights.ndim)
                        weights = weights.reshape(-1, weights.shape[-1])
                    elif weights.ndim == z.ndim:
                        # Separate, flattened weights for each model:
                        weights = weights.T if model_axis == 0 else weights
                    else:
                        # Common weights for all the models:
                        weights = weights.flatten()
            else:
                rhs = z.flatten()
                if weights is not None:
                    weights = weights.flatten()

        # If the derivative is defined along rows (as with non-linear models)
        if model_copy.col_fit_deriv:
            lhs = np.asarray(lhs).T

        # Some models (eg. Polynomial1D) don't flatten multi-dimensional inputs
        # when constructing their Vandermonde matrix, which can lead to obscure
        # failures below. Ultimately, np.linalg.lstsq can't handle >2D matrices,
        # so just raise a slightly more informative error when this happens:
        if np.asanyarray(lhs).ndim > 2:
            raise ValueError('{} gives unsupported >2D derivative matrix for '
                             'this x/y'.format(type(model_copy).__name__))

        # Subtract any terms fixed by the user from (a copy of) the RHS, in
        # order to fit the remaining terms correctly:
        if has_fixed:
            if model_copy.col_fit_deriv:
                fixderivs = np.asarray(fixderivs).T  # as for lhs above
            rhs = rhs - fixderivs.dot(fixparams)  # evaluate user-fixed terms

        # Subtract any terms implicit in the model from the RHS, which, like
        # user-fixed terms, affect the dependent variable but are not fitted:
        if sum_of_implicit_terms is not None:
            # If we have a model set, the extra axis must be added to
            # sum_of_implicit_terms as its innermost dimension, to match the
            # dimensionality of rhs after _convert_input "rolls" it as needed
            # by np.linalg.lstsq. The vector then gets broadcast to the right
            # number of sets (columns). This assumes all the models share the
            # same input coordinates, as is currently the case.
            if len(model_copy) > 1:
                sum_of_implicit_terms = sum_of_implicit_terms[..., np.newaxis]
            rhs = rhs - sum_of_implicit_terms

        if weights is not None:

            if rhs.ndim == 2:
                if weights.shape == rhs.shape:
                    # separate weights for multiple models case: broadcast
                    # lhs to have more dimension (for each model)
                    lhs = lhs[..., np.newaxis] * weights[:, np.newaxis]
                    rhs = rhs * weights
                else:
                    lhs *= weights[:, np.newaxis]
                    # Don't modify in-place in case rhs was the original
                    # dependent variable array
                    rhs = rhs * weights[:, np.newaxis]
            else:
                lhs *= weights[:, np.newaxis]
                rhs = rhs * weights

        scl = (lhs * lhs).sum(0)
        lhs /= scl

        masked = np.any(np.ma.getmask(rhs))
        if weights is not None and not masked and np.any(np.isnan(lhs)):
            raise ValueError('Found NaNs in the coefficient matrix, which '
                             'should not happen and would crash the lapack '
                             'routine. Maybe check that weights are not null.')

        a = None  # need for calculating covarience

        if ((masked and len(model_copy) > 1) or
                (weights is not None and weights.ndim > 1)):

            # Separate masks or weights for multiple models case: Numpy's
            # lstsq supports multiple dimensions only for rhs, so we need to
            # loop manually on the models. This may be fixed in the future
            # with https://github.com/numpy/numpy/pull/15777.

            # Initialize empty array of coefficients and populate it one model
            # at a time. The shape matches the number of coefficients from the
            # Vandermonde matrix and the number of models from the RHS:
            lacoef = np.zeros(lhs.shape[1:2] + rhs.shape[-1:], dtype=rhs.dtype)

            # Arrange the lhs as a stack of 2D matrices that we can iterate
            # over to get the correctly-orientated lhs for each model:
            if lhs.ndim > 2:
                lhs_stack = np.rollaxis(lhs, -1, 0)
            else:
                lhs_stack = np.broadcast_to(lhs, rhs.shape[-1:] + lhs.shape)

            # Loop over the models and solve for each one. By this point, the
            # model set axis is the second of two. Transpose rather than using,
            # say, np.moveaxis(array, -1, 0), since it's slightly faster and
            # lstsq can't handle >2D arrays anyway. This could perhaps be
            # optimized by collecting together models with identical masks
            # (eg. those with no rejected points) into one operation, though it
            # will still be relatively slow when calling lstsq repeatedly.
            for model_lhs, model_rhs, model_lacoef in zip(lhs_stack, rhs.T, lacoef.T):

                # Cull masked points on both sides of the matrix equation:
                good = ~model_rhs.mask if masked else slice(None)
                model_lhs = model_lhs[good]
                model_rhs = model_rhs[good][..., np.newaxis]
                a = model_lhs

                # Solve for this model:
                t_coef, resids, rank, sval = np.linalg.lstsq(model_lhs,
                                                             model_rhs, rcond)
                model_lacoef[:] = t_coef.T

        else:

            # If we're fitting one or more models over a common set of points,
            # we only have to solve a single matrix equation, which is an order
            # of magnitude faster than calling lstsq() once per model below:

            good = ~rhs.mask if masked else slice(None)  # latter is a no-op
            a = lhs[good]
            # Solve for one or more models:
            lacoef, resids, rank, sval = np.linalg.lstsq(lhs[good],
                                                         rhs[good], rcond)

        self.fit_info['residuals'] = resids
        self.fit_info['rank'] = rank
        self.fit_info['singular_values'] = sval

        lacoef /= scl[:, np.newaxis] if scl.ndim < rhs.ndim else scl
        self.fit_info['params'] = lacoef

        _fitter_to_model_params(model_copy, lacoef.flatten())

        # TODO: Only Polynomial models currently have an _order attribute;
        # maybe change this to read isinstance(model, PolynomialBase)
        if hasattr(model_copy, '_order') and len(model_copy) == 1 \
                and not has_fixed and rank != model_copy._order:
            warnings.warn("The fit may be poorly conditioned\n",
                          AstropyUserWarning)

        # calculate and set covariance matrix and standard devs. on model
        if self._calc_uncertainties:
            if len(y) > len(lacoef):
                self._add_fitting_uncertainties(model_copy, a*scl,
                                               len(lacoef), x, y, z, resids)
        model_copy.sync_constraints = True
        return model_copy


class FittingWithOutlierRemoval:
    """
    This class combines an outlier removal technique with a fitting procedure.
    Basically, given a maximum number of iterations ``niter``, outliers are
    removed and fitting is performed for each iteration, until no new outliers
    are found or ``niter`` is reached.

    Parameters
    ----------
    fitter : `Fitter`
        An instance of any Astropy fitter, i.e., LinearLSQFitter,
        LevMarLSQFitter, SLSQPLSQFitter, SimplexLSQFitter, JointFitter. For
        model set fitting, this must understand masked input data (as
        indicated by the fitter class attribute ``supports_masked_input``).
    outlier_func : callable
        A function for outlier removal.
        If this accepts an ``axis`` parameter like the `numpy` functions, the
        appropriate value will be supplied automatically when fitting model
        sets (unless overridden in ``outlier_kwargs``), to find outliers for
        each model separately; otherwise, the same filtering must be performed
        in a loop over models, which is almost an order of magnitude slower.
    niter : int, optional
        Maximum number of iterations.
    outlier_kwargs : dict, optional
        Keyword arguments for outlier_func.

    Attributes
    ----------
    fit_info : dict
        The ``fit_info`` (if any) from the last iteration of the wrapped
        ``fitter`` during the most recent fit. An entry is also added with the
        keyword ``niter`` that records the actual number of fitting iterations
        performed (as opposed to the user-specified maximum).
    """

    def __init__(self, fitter, outlier_func, niter=3, **outlier_kwargs):
        self.fitter = fitter
        self.outlier_func = outlier_func
        self.niter = niter
        self.outlier_kwargs = outlier_kwargs
        self.fit_info = {'niter': None}

    def __str__(self):
        return ("Fitter: {0}\nOutlier function: {1}\nNum. of iterations: {2}" +
                ("\nOutlier func. args.: {3}"))\
                .format(self.fitter.__class__.__name__,
                        self.outlier_func.__name__, self.niter,
                        self.outlier_kwargs)

    def __repr__(self):
        return ("{0}(fitter: {1}, outlier_func: {2}," +
                " niter: {3}, outlier_kwargs: {4})")\
                 .format(self.__class__.__name__,
                         self.fitter.__class__.__name__,
                         self.outlier_func.__name__, self.niter,
                         self.outlier_kwargs)

    def __call__(self, model, x, y, z=None, weights=None, **kwargs):
        """
        Parameters
        ----------
        model : `~astropy.modeling.FittableModel`
            An analytic model which will be fit to the provided data.
            This also contains the initial guess for an optimization
            algorithm.
        x : array-like
            Input coordinates.
        y : array-like
            Data measurements (1D case) or input coordinates (2D case).
        z : array-like, optional
            Data measurements (2D case).
        weights : array-like, optional
            Weights to be passed to the fitter.
        kwargs : dict, optional
            Keyword arguments to be passed to the fitter.
        Returns
        -------
        fitted_model : `~astropy.modeling.FittableModel`
            Fitted model after outlier removal.
        mask : `numpy.ndarray`
            Boolean mask array, identifying which points were used in the final
            fitting iteration (False) and which were found to be outliers or
            were masked in the input (True).
        """

        # For single models, the data get filtered here at each iteration and
        # then passed to the fitter, which is the historical behavior and
        # works even for fitters that don't understand masked arrays. For model
        # sets, the fitter must be able to filter masked data internally,
        # because fitters require a single set of x/y coordinates whereas the
        # eliminated points can vary between models. To avoid this limitation,
        # we could fall back to looping over individual model fits, but it
        # would likely be fiddly and involve even more overhead (and the
        # non-linear fitters don't work with model sets anyway, as of writing).

        if len(model) == 1:
            model_set_axis = None
        else:
            if not hasattr(self.fitter, 'supports_masked_input') or \
               self.fitter.supports_masked_input is not True:
                raise ValueError("{} cannot fit model sets with masked "
                                 "values".format(type(self.fitter).__name__))

            # Fitters use their input model's model_set_axis to determine how
            # their input data are stacked:
            model_set_axis = model.model_set_axis
        # Construct input coordinate tuples for fitters & models that are
        # appropriate for the dimensionality being fitted:
        if z is None:
            coords = (x, )
            data = y
        else:
            coords = x, y
            data = z

        # For model sets, construct a numpy-standard "axis" tuple for the
        # outlier function, to treat each model separately (if supported):
        if model_set_axis is not None:

            if model_set_axis < 0:
                model_set_axis += data.ndim

            if 'axis' not in self.outlier_kwargs:  # allow user override
                # This also works for False (like model instantiation):
                self.outlier_kwargs['axis'] = tuple(
                    n for n in range(data.ndim) if n != model_set_axis
                )

        loop = False

        # Starting fit, prior to any iteration and masking:
        fitted_model = self.fitter(model, x, y, z, weights=weights, **kwargs)
        filtered_data = np.ma.masked_array(data)
        if filtered_data.mask is np.ma.nomask:
            filtered_data.mask = False
        filtered_weights = weights
        last_n_masked = filtered_data.mask.sum()
        n = 0  # (allow recording no. of iterations when 0)

        # Perform the iterative fitting:
        for n in range(1, self.niter + 1):

            # (Re-)evaluate the last model:
            model_vals = fitted_model(*coords, model_set_axis=False)

            # Determine the outliers:
            if not loop:

                # Pass axis parameter if outlier_func accepts it, otherwise
                # prepare for looping over models:
                try:
                    filtered_data = self.outlier_func(
                        filtered_data - model_vals, **self.outlier_kwargs
                    )
                # If this happens to catch an error with a parameter other
                # than axis, the next attempt will fail accordingly:
                except TypeError:
                    if model_set_axis is None:
                        raise
                    else:
                        self.outlier_kwargs.pop('axis', None)
                        loop = True

                        # Construct MaskedArray to hold filtered values:
                        filtered_data = np.ma.masked_array(
                            filtered_data,
                            dtype=np.result_type(filtered_data, model_vals),
                            copy=True
                        )
                        # Make sure the mask is an array, not just nomask:
                        if filtered_data.mask is np.ma.nomask:
                            filtered_data.mask = False

                        # Get views transposed appropriately for iteration
                        # over the set (handling data & mask separately due to
                        # NumPy issue #8506):
                        data_T = np.rollaxis(filtered_data, model_set_axis, 0)
                        mask_T = np.rollaxis(filtered_data.mask,
                                             model_set_axis, 0)

            if loop:
                model_vals_T = np.rollaxis(model_vals, model_set_axis, 0)
                for row_data, row_mask, row_mod_vals in zip(data_T, mask_T,
                                                            model_vals_T):
                    masked_residuals = self.outlier_func(
                        row_data - row_mod_vals, **self.outlier_kwargs
                    )
                    row_data.data[:] = masked_residuals.data
                    row_mask[:] = masked_residuals.mask

                # Issue speed warning after the fact, so it only shows up when
                # the TypeError is genuinely due to the axis argument.
                warnings.warn('outlier_func did not accept axis argument; '
                              'reverted to slow loop over models.',
                              AstropyUserWarning)

            # Recombine newly-masked residuals with model to get masked values:
            filtered_data += model_vals

            # Re-fit the data after filtering, passing masked/unmasked values
            # for single models / sets, respectively:
            if model_set_axis is None:

                good = ~filtered_data.mask

                if weights is not None:
                    filtered_weights = weights[good]

                fitted_model = self.fitter(fitted_model,
                                           *(c[good] for c in coords),
                                           filtered_data.data[good],
                                           weights=filtered_weights, **kwargs)
            else:
                fitted_model = self.fitter(fitted_model, *coords,
                                           filtered_data,
                                           weights=filtered_weights, **kwargs)

            # Stop iteration if the masked points are no longer changing (with
            # cumulative rejection we only need to compare how many there are):
            this_n_masked = filtered_data.mask.sum()  # (minimal overhead)
            if this_n_masked == last_n_masked:
                break
            last_n_masked = this_n_masked

        self.fit_info = {'niter': n}
        self.fit_info.update(getattr(self.fitter, 'fit_info', {}))

        return fitted_model, filtered_data.mask


class LevMarLSQFitter(metaclass=_FitterMeta):
    """
    Levenberg-Marquardt algorithm and least squares statistic.

    Attributes
    ----------
    fit_info : dict
        The `scipy.optimize.leastsq` result for the most recent fit (see
        notes).

    Notes
    -----
    The ``fit_info`` dictionary contains the values returned by
    `scipy.optimize.leastsq` for the most recent fit, including the values from
    the ``infodict`` dictionary it returns. See the `scipy.optimize.leastsq`
    documentation for details on the meaning of these values. Note that the
    ``x`` return value is *not* included (as it is instead the parameter values
    of the returned model).
    Additionally, one additional element of ``fit_info`` is computed whenever a
    model is fit, with the key 'param_cov'. The corresponding value is the
    covariance matrix of the parameters as a 2D numpy array.  The order of the
    matrix elements matches the order of the parameters in the fitted model
    (i.e., the same order as ``model.param_names``).

    """

    supported_constraints = ['fixed', 'tied', 'bounds']
    """
    The constraint types supported by this fitter type.
    """

    def __init__(self, calc_uncertainties=False):
        self.fit_info = {'nfev': None,
                         'fvec': None,
                         'fjac': None,
                         'ipvt': None,
                         'qtf': None,
                         'message': None,
                         'ierr': None,
                         'param_jac': None,
                         'param_cov': None}
        self._calc_uncertainties=calc_uncertainties
        super().__init__()

    def objective_function(self, fps, *args):
        """
        Function to minimize.

        Parameters
        ----------
        fps : list
            parameters returned by the fitter
        args : list
            [model, [weights], [input coordinates]]

        """

        model = args[0]
        weights = args[1]
        _fitter_to_model_params(model, fps)
        meas = args[-1]
        if weights is None:
            return np.ravel(model(*args[2: -1]) - meas)
        else:
            return np.ravel(weights * (model(*args[2: -1]) - meas))

    @staticmethod
    def _add_fitting_uncertainties(model, cov_matrix):
        """
        Set ``cov_matrix`` and ``stds`` attributes on model with parameter
        covariance matrix returned by ``optimize.leastsq``.
        """

        free_param_names = [x for x in model.fixed if (model.fixed[x] is False)
                            and (model.tied[x] is False)]

        model.cov_matrix = Covariance(cov_matrix, free_param_names)
        model.stds = StandardDeviations(cov_matrix, free_param_names)

    @fitter_unit_support
    def __call__(self, model, x, y, z=None, weights=None,
                 maxiter=DEFAULT_MAXITER, acc=DEFAULT_ACC,
                 epsilon=DEFAULT_EPS, estimate_jacobian=False):
        """
        Fit data to this model.

        Parameters
        ----------
        model : `~astropy.modeling.FittableModel`
            model to fit to x, y, z
        x : array
           input coordinates
        y : array
           input coordinates
        z : array, optional
           input coordinates
        weights : array, optional
            Weights for fitting.
            For data with Gaussian uncertainties, the weights should be
            1/sigma.
        maxiter : int
            maximum number of iterations
        acc : float
            Relative error desired in the approximate solution
        epsilon : float
            A suitable step length for the forward-difference
            approximation of the Jacobian (if model.fjac=None). If
            epsfcn is less than the machine precision, it is
            assumed that the relative errors in the functions are
            of the order of the machine precision.
        estimate_jacobian : bool
            If False (default) and if the model has a fit_deriv method,
            it will be used. Otherwise the Jacobian will be estimated.
            If True, the Jacobian will be estimated in any case.
        equivalencies : list or None, optional, keyword-only
            List of *additional* equivalencies that are should be applied in
            case x, y and/or z have units. Default is None.

        Returns
        -------
        model_copy : `~astropy.modeling.FittableModel`
            a copy of the input model with parameters set by the fitter

        """

        from scipy import optimize

        model_copy = _validate_model(model, self.supported_constraints)
        model_copy.sync_constraints = False
        farg = (model_copy, weights, ) + _convert_input(x, y, z)
        if model_copy.fit_deriv is None or estimate_jacobian:
            dfunc = None
        else:
            dfunc = self._wrap_deriv
        init_values, _ = _model_to_fit_params(model_copy)
        fitparams, cov_x, dinfo, mess, ierr = optimize.leastsq(
            self.objective_function, init_values, args=farg, Dfun=dfunc,
            col_deriv=model_copy.col_fit_deriv, maxfev=maxiter, epsfcn=epsilon,
            xtol=acc, full_output=True)
        _fitter_to_model_params(model_copy, fitparams)
        self.fit_info.update(dinfo)
        self.fit_info['cov_x'] = cov_x
        self.fit_info['message'] = mess
        self.fit_info['ierr'] = ierr
        if ierr not in [1, 2, 3, 4]:
            warnings.warn("The fit may be unsuccessful; check "
                          "fit_info['message'] for more information.",
                          AstropyUserWarning)

        # now try to compute the true covariance matrix
        if (len(y) > len(init_values)) and cov_x is not None:
            sum_sqrs = np.sum(self.objective_function(fitparams, *farg)**2)
            dof = len(y) - len(init_values)
            self.fit_info['param_cov'] = cov_x * sum_sqrs / dof
        else:
            self.fit_info['param_cov'] = None

        if self._calc_uncertainties is True:
            if self.fit_info['param_cov'] is not None:
                self._add_fitting_uncertainties(model_copy,
                                               self.fit_info['param_cov'])

        model_copy.sync_constraints = True
        return model_copy

    @staticmethod
    def _wrap_deriv(params, model, weights, x, y, z=None):
        """
        Wraps the method calculating the Jacobian of the function to account
        for model constraints.
        `scipy.optimize.leastsq` expects the function derivative to have the
        above signature (parlist, (argtuple)). In order to accommodate model
        constraints, instead of using p directly, we set the parameter list in
        this function.
        """

        if weights is None:
            weights = 1.0

        if any(model.fixed.values()) or any(model.tied.values()):
            # update the parameters with the current values from the fitter
            _fitter_to_model_params(model, params)
            if z is None:
                full = np.array(model.fit_deriv(x, *model.parameters))
                if not model.col_fit_deriv:
                    full_deriv = np.ravel(weights) * full.T
                else:
                    full_deriv = np.ravel(weights) * full
            else:
                full = np.array([np.ravel(_) for _ in model.fit_deriv(x, y, *model.parameters)])
                if not model.col_fit_deriv:
                    full_deriv = np.ravel(weights) * full.T
                else:
                    full_deriv = np.ravel(weights) * full

            pars = [getattr(model, name) for name in model.param_names]
            fixed = [par.fixed for par in pars]
            tied = [par.tied for par in pars]
            tied = list(np.where([par.tied is not False for par in pars],
                                 True, tied))
            fix_and_tie = np.logical_or(fixed, tied)
            ind = np.logical_not(fix_and_tie)

            if not model.col_fit_deriv:
                residues = np.asarray(full_deriv[np.nonzero(ind)]).T
            else:
                residues = full_deriv[np.nonzero(ind)]

            return [np.ravel(_) for _ in residues]
        else:
            if z is None:
                try:
                    return np.array([np.ravel(_) for _ in np.array(weights) *
                                     np.array(model.fit_deriv(x, *params))])
                except ValueError:
                    return np.array([np.ravel(_) for _ in np.array(weights) *
                                     np.moveaxis(
                                         np.array(model.fit_deriv(x, *params)),
                                         -1, 0)]).transpose()
            else:
                if not model.col_fit_deriv:
                    return [np.ravel(_) for _ in
                            (np.ravel(weights) * np.array(model.fit_deriv(x, y, *params)).T).T]
                return [np.ravel(_) for _ in weights * np.array(model.fit_deriv(x, y, *params))]


class SLSQPLSQFitter(Fitter):
    """
    Sequential Least Squares Programming (SLSQP) optimization algorithm and
    least squares statistic.

    Raises
    ------
    ModelLinearityError
        A linear model is passed to a nonlinear fitter

    Notes
    ------
    See also the `~astropy.modeling.optimizers.SLSQP` optimizer.

    """

    supported_constraints = SLSQP.supported_constraints

    def __init__(self):
        super().__init__(optimizer=SLSQP, statistic=leastsquare)
        self.fit_info = {}

    @fitter_unit_support
    def __call__(self, model, x, y, z=None, weights=None, **kwargs):
        """
        Fit data to this model.

        Parameters
        ----------
        model : `~astropy.modeling.FittableModel`
            model to fit to x, y, z
        x : array
            input coordinates
        y : array
            input coordinates
        z : array, optional
            input coordinates
        weights : array, optional
            Weights for fitting.
            For data with Gaussian uncertainties, the weights should be
            1/sigma.
        kwargs : dict
            optional keyword arguments to be passed to the optimizer or the statistic
        verblevel : int
            0-silent
            1-print summary upon completion,
            2-print summary after each iteration
        maxiter : int
            maximum number of iterations
        epsilon : float
            the step size for finite-difference derivative estimates
        acc : float
            Requested accuracy
        equivalencies : list or None, optional, keyword-only
            List of *additional* equivalencies that are should be applied in
            case x, y and/or z have units. Default is None.

        Returns
        -------
        model_copy : `~astropy.modeling.FittableModel`
            a copy of the input model with parameters set by the fitter

        """

        model_copy = _validate_model(model, self._opt_method.supported_constraints)
        model_copy.sync_constraints = False
        farg = _convert_input(x, y, z)
        farg = (model_copy, weights, ) + farg
        init_values, _ = _model_to_fit_params(model_copy)
        fitparams, self.fit_info = self._opt_method(
            self.objective_function, init_values, farg, **kwargs)
        _fitter_to_model_params(model_copy, fitparams)

        model_copy.sync_constraints = True
        return model_copy


class SimplexLSQFitter(Fitter):
    """
    Simplex algorithm and least squares statistic.

    Raises
    ------
    `ModelLinearityError`
        A linear model is passed to a nonlinear fitter

    """

    supported_constraints = Simplex.supported_constraints

    def __init__(self):
        super().__init__(optimizer=Simplex, statistic=leastsquare)
        self.fit_info = {}

    @fitter_unit_support
    def __call__(self, model, x, y, z=None, weights=None, **kwargs):
        """
        Fit data to this model.

        Parameters
        ----------
        model : `~astropy.modeling.FittableModel`
            model to fit to x, y, z
        x : array
            input coordinates
        y : array
            input coordinates
        z : array, optional
            input coordinates
        weights : array, optional
            Weights for fitting.
            For data with Gaussian uncertainties, the weights should be
            1/sigma.
        kwargs : dict
            optional keyword arguments to be passed to the optimizer or the statistic
        maxiter : int
            maximum number of iterations
        acc : float
            Relative error in approximate solution
        equivalencies : list or None, optional, keyword-only
            List of *additional* equivalencies that are should be applied in
            case x, y and/or z have units. Default is None.

        Returns
        -------
        model_copy : `~astropy.modeling.FittableModel`
            a copy of the input model with parameters set by the fitter

        """

        model_copy = _validate_model(model,
                                     self._opt_method.supported_constraints)
        model_copy.sync_constraints = False
        farg = _convert_input(x, y, z)
        farg = (model_copy, weights, ) + farg

        init_values, _ = _model_to_fit_params(model_copy)

        fitparams, self.fit_info = self._opt_method(
            self.objective_function, init_values, farg, **kwargs)
        _fitter_to_model_params(model_copy, fitparams)
        model_copy.sync_constraints = True
        return model_copy


class JointFitter(metaclass=_FitterMeta):
    """
    Fit models which share a parameter.
    For example, fit two gaussians to two data sets but keep
    the FWHM the same.

    Parameters
    ----------
    models : list
        a list of model instances
    jointparameters : list
        a list of joint parameters
    initvals : list
        a list of initial values

    """

    def __init__(self, models, jointparameters, initvals):
        self.models = list(models)
        self.initvals = list(initvals)
        self.jointparams = jointparameters
        self._verify_input()
        self.fitparams = self._model_to_fit_params()

        # a list of model.n_inputs
        self.modeldims = [m.n_inputs for m in self.models]
        # sum all model dimensions
        self.ndim = np.sum(self.modeldims)

    def _model_to_fit_params(self):
        fparams = []
        fparams.extend(self.initvals)
        for model in self.models:
            params = model.parameters.tolist()
            joint_params = self.jointparams[model]
            param_metrics = model._param_metrics
            for param_name in joint_params:
                slice_ = param_metrics[param_name]['slice']
                del params[slice_]
            fparams.extend(params)
        return fparams

    def objective_function(self, fps, *args):
        """
        Function to minimize.

        Parameters
        ----------
        fps : list
            the fitted parameters - result of an one iteration of the
            fitting algorithm
        args : dict
            tuple of measured and input coordinates
            args is always passed as a tuple from optimize.leastsq

        """

        lstsqargs = list(args)
        fitted = []
        fitparams = list(fps)
        numjp = len(self.initvals)
        # make a separate list of the joint fitted parameters
        jointfitparams = fitparams[:numjp]
        del fitparams[:numjp]

        for model in self.models:
            joint_params = self.jointparams[model]
            margs = lstsqargs[:model.n_inputs + 1]
            del lstsqargs[:model.n_inputs + 1]
            # separate each model separately fitted parameters
            numfp = len(model._parameters) - len(joint_params)
            mfparams = fitparams[:numfp]

            del fitparams[:numfp]
            # recreate the model parameters
            mparams = []
            param_metrics = model._param_metrics
            for param_name in model.param_names:
                if param_name in joint_params:
                    index = joint_params.index(param_name)
                    # should do this with slices in case the
                    # parameter is not a number
                    mparams.extend([jointfitparams[index]])
                else:
                    slice_ = param_metrics[param_name]['slice']
                    plen = slice_.stop - slice_.start
                    mparams.extend(mfparams[:plen])
                    del mfparams[:plen]
            modelfit = model.evaluate(margs[:-1], *mparams)
            fitted.extend(modelfit - margs[-1])
        return np.ravel(fitted)

    def _verify_input(self):
        if len(self.models) <= 1:
            raise TypeError(f"Expected >1 models, {len(self.models)} is given")
        if len(self.jointparams.keys()) < 2:
            raise TypeError("At least two parameters are expected, "
                            "{} is given".format(len(self.jointparams.keys())))
        for j in self.jointparams.keys():
            if len(self.jointparams[j]) != len(self.initvals):
                raise TypeError("{} parameter(s) provided but {} expected".format(
                    len(self.jointparams[j]), len(self.initvals)))

    def __call__(self, *args):
        """
        Fit data to these models keeping some of the parameters common to the
        two models.
        """

        from scipy import optimize

        if len(args) != reduce(lambda x, y: x + 1 + y + 1, self.modeldims):
            raise ValueError("Expected {} coordinates in args but {} provided"
                             .format(reduce(lambda x, y: x + 1 + y + 1,
                                            self.modeldims), len(args)))

        self.fitparams[:], _ = optimize.leastsq(self.objective_function,
                                                self.fitparams, args=args)

        fparams = self.fitparams[:]
        numjp = len(self.initvals)
        # make a separate list of the joint fitted parameters
        jointfitparams = fparams[:numjp]
        del fparams[:numjp]

        for model in self.models:
            # extract each model's fitted parameters
            joint_params = self.jointparams[model]
            numfp = len(model._parameters) - len(joint_params)
            mfparams = fparams[:numfp]

            del fparams[:numfp]
            # recreate the model parameters
            mparams = []
            param_metrics = model._param_metrics
            for param_name in model.param_names:
                if param_name in joint_params:
                    index = joint_params.index(param_name)
                    # should do this with slices in case the parameter
                    # is not a number
                    mparams.extend([jointfitparams[index]])
                else:
                    slice_ = param_metrics[param_name]['slice']
                    plen = slice_.stop - slice_.start
                    mparams.extend(mfparams[:plen])
                    del mfparams[:plen]
            model.parameters = np.array(mparams)


def _convert_input(x, y, z=None, n_models=1, model_set_axis=0):
    """Convert inputs to float arrays."""

    x = np.asanyarray(x, dtype=float)
    y = np.asanyarray(y, dtype=float)

    if z is not None:
        z = np.asanyarray(z, dtype=float)
        data_ndim, data_shape = z.ndim, z.shape
    else:
        data_ndim, data_shape = y.ndim, y.shape

    # For compatibility with how the linear fitter code currently expects to
    # work, shift the dependent variable's axes to the expected locations
    if n_models > 1 or data_ndim > x.ndim:
        if (model_set_axis or 0) >= data_ndim:
            raise ValueError("model_set_axis out of range")
        if data_shape[model_set_axis] != n_models:
            raise ValueError(
                "Number of data sets (y or z array) is expected to equal "
                "the number of parameter sets"
            )
        if z is None:
            # For a 1-D model the y coordinate's model-set-axis is expected to
            # be last, so that its first dimension is the same length as the x
            # coordinates.  This is in line with the expectations of
            # numpy.linalg.lstsq:
            # https://numpy.org/doc/stable/reference/generated/numpy.linalg.lstsq.html
            # That is, each model should be represented by a column.  TODO:
            # Obviously this is a detail of np.linalg.lstsq and should be
            # handled specifically by any fitters that use it...
            y = np.rollaxis(y, model_set_axis, y.ndim)
            data_shape = y.shape[:-1]
        else:
            # Shape of z excluding model_set_axis
            data_shape = (z.shape[:model_set_axis] +
                          z.shape[model_set_axis + 1:])

    if z is None:
        if data_shape != x.shape:
            raise ValueError("x and y should have the same shape")
        farg = (x, y)
    else:
        if not (x.shape == y.shape == data_shape):
            raise ValueError("x, y and z should have the same shape")
        farg = (x, y, z)
    return farg


# TODO: These utility functions are really particular to handling
# bounds/tied/fixed constraints for scipy.optimize optimizers that do not
# support them inherently; this needs to be reworked to be clear about this
# distinction (and the fact that these are not necessarily applicable to any
# arbitrary fitter--as evidenced for example by the fact that JointFitter has
# its own versions of these)
# TODO: Most of this code should be entirely rewritten; it should not be as
# inefficient as it is.
def _fitter_to_model_params(model, fps):
    """
    Constructs the full list of model parameters from the fitted and
    constrained parameters.
    """

    _, fit_param_indices = _model_to_fit_params(model)

    has_tied = any(model.tied.values())
    has_fixed = any(model.fixed.values())
    has_bound = any(b != (None, None) for b in model.bounds.values())
    parameters = model.parameters

    if not (has_tied or has_fixed or has_bound):
        # We can just assign directly
        model.parameters = fps
        return

    fit_param_indices = set(fit_param_indices)
    offset = 0
    param_metrics = model._param_metrics
    for idx, name in enumerate(model.param_names):
        if idx not in fit_param_indices:
            continue

        slice_ = param_metrics[name]['slice']
        shape = param_metrics[name]['shape']
        # This is determining which range of fps (the fitted parameters) maps
        # to parameters of the model
        size = reduce(operator.mul, shape, 1)

        values = fps[offset:offset + size]

        # Check bounds constraints
        if model.bounds[name] != (None, None):
            _min, _max = model.bounds[name]
            if _min is not None:
                values = np.fmax(values, _min)
            if _max is not None:
                values = np.fmin(values, _max)

        parameters[slice_] = values
        offset += size

    # Update model parameters before calling ``tied`` constraints.
    model._array_to_parameters()

    # This has to be done in a separate loop due to how tied parameters are
    # currently evaluated (the fitted parameters need to actually be *set* on
    # the model first, for use in evaluating the "tied" expression--it might be
    # better to change this at some point
    if has_tied:
        for idx, name in enumerate(model.param_names):
            if model.tied[name]:
                value = model.tied[name](model)
                slice_ = param_metrics[name]['slice']

                # To handle multiple tied constraints, model parameters
                # need to be updated after each iteration.
                parameters[slice_] = value
                model._array_to_parameters()


def _model_to_fit_params(model):
    """
    Convert a model instance's parameter array to an array that can be used
    with a fitter that doesn't natively support fixed or tied parameters.
    In particular, it removes fixed/tied parameters from the parameter
    array.
    These may be a subset of the model parameters, if some of them are held
    constant or tied.
    """

    fitparam_indices = list(range(len(model.param_names)))
    if any(model.fixed.values()) or any(model.tied.values()):
        params = list(model.parameters)
        param_metrics = model._param_metrics
        for idx, name in list(enumerate(model.param_names))[::-1]:
            if model.fixed[name] or model.tied[name]:
                slice_ = param_metrics[name]['slice']
                del params[slice_]
                del fitparam_indices[idx]
        return (np.array(params), fitparam_indices)
    return (model.parameters, fitparam_indices)


def _validate_constraints(supported_constraints, model):
    """Make sure model constraints are supported by the current fitter."""

    message = 'Optimizer cannot handle {0} constraints.'

    if (any(model.fixed.values()) and
            'fixed' not in supported_constraints):
        raise UnsupportedConstraintError(
            message.format('fixed parameter'))

    if any(model.tied.values()) and 'tied' not in supported_constraints:
        raise UnsupportedConstraintError(
            message.format('tied parameter'))

    if (any(tuple(b) != (None, None) for b in model.bounds.values()) and
            'bounds' not in supported_constraints):
        raise UnsupportedConstraintError(
            message.format('bound parameter'))

    if model.eqcons and 'eqcons' not in supported_constraints:
        raise UnsupportedConstraintError(message.format('equality'))

    if model.ineqcons and 'ineqcons' not in supported_constraints:
        raise UnsupportedConstraintError(message.format('inequality'))


def _validate_model(model, supported_constraints):
    """
    Check that model and fitter are compatible and return a copy of the model.
    """

    if not model.fittable:
        raise ValueError("Model does not appear to be fittable.")
    if model.linear:
        warnings.warn('Model is linear in parameters; '
                      'consider using linear fitting methods.',
                      AstropyUserWarning)
    elif len(model) != 1:
        # for now only single data sets ca be fitted
        raise ValueError("Non-linear fitters can only fit "
                         "one data set at a time.")
    _validate_constraints(supported_constraints, model)

    model_copy = model.copy()
    return model_copy


def populate_entry_points(entry_points):
    """
    This injects entry points into the `astropy.modeling.fitting` namespace.
    This provides a means of inserting a fitting routine without requirement
    of it being merged into astropy's core.

    Parameters
    ----------
    entry_points : list of `~importlib.metadata.EntryPoint`
        entry_points are objects which encapsulate importable objects and
        are defined on the installation of a package.

    Notes
    -----
    An explanation of entry points can be found `here <http://setuptools.readthedocs.io/en/latest/setuptools.html#dynamic-discovery-of-services-and-plugins>`
    """

    for entry_point in entry_points:
        name = entry_point.name
        try:
            entry_point = entry_point.load()
        except Exception as e:
            # This stops the fitting from choking if an entry_point produces an error.
            warnings.warn(AstropyUserWarning(
                f'{type(e).__name__} error occurred in entry point {name}.'))
        else:
            if not inspect.isclass(entry_point):
                warnings.warn(AstropyUserWarning(
                    f'Modeling entry point {name} expected to be a Class.'))
            else:
                if issubclass(entry_point, Fitter):
                    name = entry_point.__name__
                    globals()[name] = entry_point
                    __all__.append(name)
                else:
                    warnings.warn(AstropyUserWarning(
                        'Modeling entry point {} expected to extend '
                        'astropy.modeling.Fitter' .format(name)))


def _populate_ep():
    # TODO: Exclusively use select when Python minversion is 3.10
    ep = entry_points()
    if hasattr(ep, 'select'):
        populate_entry_points(ep.select(group='astropy.modeling'))
    else:
        populate_entry_points(ep.get('astropy.modeling', []))


_populate_ep()



import numpy as np
from numpy.testing import assert_allclose

import pytest
from pytest import raises

from astropy import units as u

from astropy.wcs import WCS
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs.wcsapi.utils import deserialize_class, wcs_info_str


def test_construct():

    result = deserialize_class(('astropy.units.Quantity', (10,), {'unit': 'deg'}))
    assert_quantity_allclose(result, 10 * u.deg)


def test_noconstruct():

    result = deserialize_class(('astropy.units.Quantity', (), {'unit': 'deg'}), construct=False)
    assert result == (u.Quantity, (), {'unit': 'deg'})


def test_invalid():

    with raises(ValueError) as exc:
        deserialize_class(('astropy.units.Quantity', (), {'unit': 'deg'}, ()))
    assert exc.value.args[0] == 'Expected a tuple of three values'


DEFAULT_1D_STR = """
WCS Transformation

This transformation has 1 pixel and 1 world dimensions

Array shape (Numpy order): None

Pixel Dim  Axis Name  Data size  Bounds
        0  None            None  None

World Dim  Axis Name  Physical Type  Units
        0  None       None           unknown

Correlation between pixel and world axes:

           Pixel Dim
World Dim    0
        0  yes
"""


def test_wcs_info_str():

    # The tests in test_sliced_low_level_wcs.py exercise wcs_info_str
    # extensively. This test is to ensure that the function exists and the
    # API of the function works as expected.

    wcs_empty = WCS(naxis=1)

    assert wcs_info_str(wcs_empty).strip() == DEFAULT_1D_STR.strip()



# -*- coding: utf-8 -*-
# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Test replacements for ERFA functions atciqz and aticq."""

import pytest
import erfa

from astropy.tests.helper import assert_quantity_allclose as assert_allclose
from astropy.time import Time
import astropy.units as u
from astropy.coordinates.builtin_frames.utils import get_jd12, atciqz, aticq
from astropy.coordinates import SphericalRepresentation

# Hard-coded random values
sph = SphericalRepresentation(lon=[15., 214.] * u.deg,
                              lat=[-12., 64.] * u.deg,
                              distance=[1, 1.])


@pytest.mark.parametrize('t', [Time("2014-06-25T00:00"),
                               Time(["2014-06-25T00:00", "2014-09-24"])])
@pytest.mark.parametrize('pos', [sph[0], sph])
def test_atciqz_aticq(t, pos):
    """Check replacements against erfa versions for consistency."""
    jd1, jd2 = get_jd12(t, 'tdb')
    astrom, _ = erfa.apci13(jd1, jd2)

    ra = pos.lon.to_value(u.rad)
    dec = pos.lat.to_value(u.rad)
    assert_allclose(erfa.atciqz(ra, dec, astrom), atciqz(pos, astrom))
    assert_allclose(erfa.aticq(ra, dec, astrom), aticq(pos, astrom))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
# -*- coding: utf-8 -*-

from astropy.coordinates import Angle, Latitude, Longitude

from astropy.io.misc.asdf.tags.unit.quantity import QuantityType


__all__ = ['AngleType', 'LatitudeType', 'LongitudeType']


class AngleType(QuantityType):
    name = "coordinates/angle"
    types = [Angle]
    requires = ['astropy']
    version = "1.0.0"
    organization = 'astropy.org'
    standard = 'astropy'

    @classmethod
    def from_tree(cls, node, ctx):
        return Angle(super().from_tree(node, ctx))


class LatitudeType(AngleType):
    name = "coordinates/latitude"
    types = [Latitude]

    @classmethod
    def from_tree(cls, node, ctx):
        return Latitude(super().from_tree(node, ctx))


class LongitudeType(AngleType):
    name = "coordinates/longitude"
    types = [Longitude]

    @classmethod
    def from_tree(cls, node, ctx):
        wrap_angle = node['wrap_angle']
        return Longitude(super().from_tree(node, ctx), wrap_angle=wrap_angle)

    @classmethod
    def to_tree(cls, longitude, ctx):
        tree = super().to_tree(longitude, ctx)
        tree['wrap_angle'] = longitude.wrap_angle

        return tree



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

from astropy.cosmology import parameters, realizations
from astropy.cosmology.realizations import Planck13, default_cosmology
from astropy.utils.exceptions import AstropyDeprecationWarning

cosmo_realz = [s for s in parameters.available if s != "Planck18_arXiv_v2"]


class Test_default_cosmology(object):
    """Tests for :class:`~astropy.cosmology.realizations.default_cosmology`."""

    @staticmethod
    def test_get_cosmology_from_string():
        """Test method ``get_cosmology_from_string``."""
        cosmo = default_cosmology.get_cosmology_from_string("no_default")
        assert cosmo is None

        cosmo = default_cosmology.get_cosmology_from_string("Planck13")
        assert cosmo is Planck13

        with pytest.raises(ValueError):
            cosmo = default_cosmology.get_cosmology_from_string("fail!")

    @staticmethod
    def test_validate_specific():
        """Test method ``validate`` for specific values."""
        value = default_cosmology.validate(None)
        assert value is realizations.Planck18

        with pytest.warns(AstropyDeprecationWarning) as e:
            value = default_cosmology.validate("Planck18_arXiv_v2")
        assert "use Planck18 instead" in str(e.list[0].message)
        assert value is realizations.Planck18_arXiv_v2

        with pytest.raises(TypeError) as e:
            default_cosmology.validate(TypeError)
        assert "string or Cosmology instance" in str(e.value)

    @pytest.mark.parametrize("name", cosmo_realz)
    def test_validate_str(self, name):
        """Test method ``validate`` for string input."""
        value = default_cosmology.validate(name)
        assert value is getattr(realizations, name)

    @pytest.mark.parametrize("name", cosmo_realz)
    def test_validate_cosmo(self, name):
        """Test method ``validate`` for cosmology instance input."""
        cosmo = getattr(realizations, name)
        value = default_cosmology.validate(cosmo)
        assert value is cosmo



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest

import numpy as np

from astropy import wcs


def test_wtbarr_i(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].i == 1


def test_wtbarr_m(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].m == 1


def test_wtbarr_m(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].kind == 'c'


def test_wtbarr_extnam(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].extnam == 'WCS-TABLE'


def test_wtbarr_extver(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].extver == 1


def test_wtbarr_extlev(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].extlev == 1


def test_wtbarr_ttype(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].ttype == 'wavelength'


def test_wtbarr_row(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].row == 1


def test_wtbarr_ndim(tab_wcs_2di):
    assert tab_wcs_2di.wcs.wtb[0].ndim == 3


def test_wtbarr_print(tab_wcs_2di, capfd):
    tab_wcs_2di.wcs.wtb[0].print_contents()
    captured = capfd.readouterr()
    s = str(tab_wcs_2di.wcs.wtb[0])
    lines = s.split('\n')
    assert captured.out == s
    assert '     i: 1' == lines[0]
    assert '     m: 1' == lines[1]
    assert '  kind: c' == lines[2]
    assert 'extnam: WCS-TABLE' == lines[3]
    assert 'extver: 1' == lines[4]
    assert 'extlev: 1' == lines[5]
    assert ' ttype: wavelength' == lines[6]
    assert '   row: 1' == lines[7]
    assert '  ndim: 3' == lines[8]
    assert lines[9].startswith('dimlen: ')
    assert '        0:   4' == lines[10]
    assert '        1:   2' == lines[11]
    assert lines[12].startswith('arrayp: ')



import pytest
import numpy as np

from astropy.coordinates import SkyCoord
from astropy.units import Quantity
from astropy.wcs import WCS
from astropy.wcs.wcsapi import BaseLowLevelWCS

# NOTE: This module is deprecated and is emitting warning.
collect_ignore = ['sliced_low_level_wcs.py']


@pytest.fixture
def spectral_1d_fitswcs():
    wcs = WCS(naxis=1)
    wcs.wcs.ctype = 'FREQ',
    wcs.wcs.cunit = 'Hz',
    wcs.wcs.cdelt = 3.e9,
    wcs.wcs.crval = 4.e9,
    wcs.wcs.crpix = 11.,
    wcs.wcs.cname = 'Frequency',
    return wcs


@pytest.fixture
def time_1d_fitswcs():
    wcs = WCS(naxis=1)
    wcs.wcs.ctype = 'TIME',
    wcs.wcs.mjdref = (30042, 0)
    wcs.wcs.crval = 3.,
    wcs.wcs.crpix = 11.,
    wcs.wcs.cname = 'Time',
    wcs.wcs.cunit = 's'
    return wcs


@pytest.fixture
def celestial_2d_fitswcs():
    wcs = WCS(naxis=2)
    wcs.wcs.ctype = 'RA---CAR', 'DEC--CAR'
    wcs.wcs.cunit = 'deg', 'deg'
    wcs.wcs.cdelt = -2., 2.
    wcs.wcs.crval = 4., 0.
    wcs.wcs.crpix = 6., 7.
    wcs.wcs.cname = 'Right Ascension', 'Declination'
    wcs.pixel_shape = (6, 7)
    wcs.pixel_bounds = [(-1, 5), (1, 7)]
    return wcs


@pytest.fixture
def spectral_cube_3d_fitswcs():
    wcs = WCS(naxis=3)
    wcs.wcs.ctype = 'RA---CAR', 'DEC--CAR', 'FREQ'
    wcs.wcs.cunit = 'deg', 'deg', 'Hz'
    wcs.wcs.cdelt = -2., 2., 3.e9
    wcs.wcs.crval = 4., 0., 4.e9
    wcs.wcs.crpix = 6., 7., 11.
    wcs.wcs.cname = 'Right Ascension', 'Declination', 'Frequency'
    wcs.pixel_shape = (6, 7, 3)
    wcs.pixel_bounds = [(-1, 5), (1, 7), (1, 2.5)]
    return wcs


@pytest.fixture
def cube_4d_fitswcs():
    wcs = WCS(naxis=4)
    wcs.wcs.ctype = 'RA---CAR', 'DEC--CAR', 'FREQ', 'TIME'
    wcs.wcs.cunit = 'deg', 'deg', 'Hz', 's'
    wcs.wcs.cdelt = -2., 2., 3.e9, 1
    wcs.wcs.crval = 4., 0., 4.e9, 3,
    wcs.wcs.crpix = 6., 7., 11., 11.
    wcs.wcs.cname = 'Right Ascension', 'Declination', 'Frequency', 'Time'
    wcs.wcs.mjdref = (30042, 0)
    return wcs


class Spectral1DLowLevelWCS(BaseLowLevelWCS):

    @property
    def pixel_n_dim(self):
        return 1

    @property
    def world_n_dim(self):
        return 1

    @property
    def world_axis_physical_types(self):
        return 'em.freq',

    @property
    def world_axis_units(self):
        return 'Hz',

    @property
    def world_axis_names(self):
        return 'Frequency',

    _pixel_shape = None

    @property
    def pixel_shape(self):
        return self._pixel_shape

    @pixel_shape.setter
    def pixel_shape(self, value):
        self._pixel_shape = value

    _pixel_bounds = None

    @property
    def pixel_bounds(self):
        return self._pixel_bounds

    @pixel_bounds.setter
    def pixel_bounds(self, value):
        self._pixel_bounds = value

    def pixel_to_world_values(self, pixel_array):
        return np.asarray(pixel_array - 10) * 3e9 + 4e9

    def world_to_pixel_values(self, world_array):
        return np.asarray(world_array - 4e9) / 3e9 + 10

    @property
    def world_axis_object_components(self):
        return ('test', 0, 'value'),

    @property
    def world_axis_object_classes(self):
        return {'test': (Quantity, (), {'unit': 'Hz'})}


@pytest.fixture
def spectral_1d_ape14_wcs():
    return Spectral1DLowLevelWCS()


class Celestial2DLowLevelWCS(BaseLowLevelWCS):

    @property
    def pixel_n_dim(self):
        return 2

    @property
    def world_n_dim(self):
        return 2

    @property
    def world_axis_physical_types(self):
        return 'pos.eq.ra', 'pos.eq.dec'

    @property
    def world_axis_units(self):
        return 'deg', 'deg'

    @property
    def world_axis_names(self):
        return 'Right Ascension', 'Declination'

    @property
    def pixel_shape(self):
        return (6, 7)

    @property
    def pixel_bounds(self):
        return (-1, 5), (1, 7)

    def pixel_to_world_values(self, px, py):
        return (-(np.asarray(px) - 5.) * 2 + 4.,
                (np.asarray(py) - 6.) * 2)

    def world_to_pixel_values(self, wx, wy):
        return (-(np.asarray(wx) - 4.) / 2 + 5.,
                np.asarray(wy) / 2 + 6.)

    @property
    def world_axis_object_components(self):
        return [('test', 0, 'spherical.lon.degree'),
                ('test', 1, 'spherical.lat.degree')]

    @property
    def world_axis_object_classes(self):
        return {'test': (SkyCoord, (), {'unit': 'deg'})}


@pytest.fixture
def celestial_2d_ape14_wcs():
    return Celestial2DLowLevelWCS()



import dask.array as da

from astropy.utils.data_info import ParentDtypeInfo

__all__ = ['as_dask_column']


class DaskInfo(ParentDtypeInfo):
    @staticmethod
    def default_format(val):
        return f'{val.compute()}'


class DaskColumn(da.Array):

    info = DaskInfo()

    def copy(self):
        # Array hard-codes the resulting copied array as Array, so need to
        # overload this since Table tries to copy the array.
        return as_dask_column(self, info=self.info)

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if isinstance(item, int):
            return result
        else:
            return as_dask_column(result, info=self.info)

    def insert(self, obj, values, axis=0):
        return as_dask_column(da.insert(self, obj, values, axis=axis),
                              info=self.info)


def as_dask_column(array, info=None):
    result = DaskColumn(array.dask, array.name, array.chunks, meta=array)
    if info is not None:
        result.info = info
    return result



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pickle

import pytest

from astropy import cosmology
from astropy.cosmology import FLRW
from astropy.tests.helper import check_pickling_recovery, pickle_protocol

originals = [FLRW]
xfails = [False]


@pytest.mark.parametrize(("original", "xfail"), zip(originals, xfails))
def test_flrw(pickle_protocol, original, xfail):
    if xfail:
        pytest.xfail()
    check_pickling_recovery(original, pickle_protocol)


@pytest.mark.parametrize("name", cosmology.parameters.available)
@pytest.mark.parametrize("protocol", [0, 1, 2, 3, 4])  # add [5] when drop 3.7
def test_builtin_realizations(name, protocol):
    """
    Test in-built realizations can pickle and unpickle.
    Also a regression test for #12008.
    """
    # get class instance
    original = getattr(cosmology, name)

    # pickle and unpickle
    f = pickle.dumps(original, protocol=protocol)
    unpickled = pickle.loads(f)

    # test equality
    assert unpickled == original
    assert unpickled.meta == original.meta



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from os.path import join
from collections import defaultdict

from setuptools import Extension

from extension_helpers import import_file

wcs_setup_package = import_file(join('astropy', 'wcs', 'setup_package.py'))


MODELING_ROOT = os.path.relpath(os.path.dirname(__file__))
MODELING_SRC = join(MODELING_ROOT, 'src')
SRC_FILES = [join(MODELING_SRC, 'projections.c.templ'),
             __file__]
GEN_FILES = [join(MODELING_SRC, 'projections.c')]


# This defines the set of projection functions that we want to wrap.
# The key is the projection name, and the value is the number of
# parameters.

# (These are in the order that the appear in the WCS coordinate
# systems paper).
projections = {
    'azp': 2,
    'szp': 3,
    'tan': 0,
    'stg': 0,
    'sin': 2,
    'arc': 0,
    'zea': 0,
    'air': 1,
    'cyp': 2,
    'cea': 1,
    'mer': 0,
    'sfl': 0,
    'par': 0,
    'mol': 0,
    'ait': 0,
    'cop': 2,
    'coe': 2,
    'cod': 2,
    'coo': 2,
    'bon': 1,
    'pco': 0,
    'tsc': 0,
    'csc': 0,
    'qsc': 0,
    'hpx': 2,
    'xph': 0,
}


def get_extensions():

    from jinja2 import Environment, FileSystemLoader

    # Prepare the jinja2 templating environment
    env = Environment(loader=FileSystemLoader(MODELING_SRC))

    c_in = env.get_template('projections.c.templ')
    c_out = c_in.render(projections=projections)

    with open(join(MODELING_SRC, 'projections.c'), 'w') as fd:
        fd.write(c_out)

    wcslib_files = [  # List of wcslib files to compile
        'prj.c',
        'wcserr.c',
        'wcsprintf.c',
        'wcsutil.c'
    ]

    wcslib_config_paths = [
        join(MODELING_SRC, 'wcsconfig.h')
    ]

    cfg = defaultdict(list)

    wcs_setup_package.get_wcslib_cfg(cfg, wcslib_files, wcslib_config_paths)

    cfg['include_dirs'].append(MODELING_SRC)

    astropy_files = [  # List of astropy.modeling files to compile
        'projections.c'
    ]
    cfg['sources'].extend(join(MODELING_SRC, x) for x in astropy_files)

    cfg['sources'] = [str(x) for x in cfg['sources']]
    cfg = dict((str(key), val) for key, val in cfg.items())

    return [Extension('astropy.modeling._projections', **cfg)]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Convolution Model"""
# pylint: disable=line-too-long, too-many-lines, too-many-arguments, invalid-name
import numpy as np

from .core import CompoundModel, SPECIAL_OPERATORS


class Convolution(CompoundModel):
    """
    Wrapper class for a convolution model.

    Parameters
    ----------
    operator: tuple
        The SPECIAL_OPERATORS entry for the convolution being used.
    model : Model
        The model for the convolution.
    kernel: Model
        The kernel model for the convolution.
    bounding_box : tuple
        A bounding box to define the limits of the integration
        approximation for the convolution.
    resolution : float
        The resolution for the approximation of the convolution.
    cache : bool, optional
        Allow convolution computation to be cached for reuse. This is
        enabled by default.

    Notes
    -----
    This is wrapper is necessary to handle the limitations of the
    pseudospectral convolution binary operator implemented in
    astropy.convolution under `~astropy.convolution.convolve_fft`. In this `~astropy.convolution.convolve_fft` it
    is assumed that the inputs ``array`` and ``kernel`` span a sufficient
    portion of the support of the functions of the convolution.
    Consequently, the ``Compound`` created by the `~astropy.convolution.convolve_models` function
    makes the assumption that one should pass an input array that
    sufficiently spans this space. This means that slightly different
    input arrays to this model will result in different outputs, even
    on points of intersection between these arrays.

    This issue is solved by requiring a ``bounding_box`` together with a
    resolution so that one can pre-calculate the entire domain and then
    (by default) cache the convolution values. The function then just
    interpolates the results from this cache.
    """

    def __init__(self, operator, model, kernel, bounding_box, resolution, cache=True):
        super().__init__(operator, model, kernel)

        self.bounding_box = bounding_box
        self._resolution = resolution

        self._cache_convolution = cache
        self._kwargs = None
        self._convolution = None

    def clear_cache(self):
        """
        Clears the cached convolution
        """

        self._kwargs = None
        self._convolution = None

    def _get_convolution(self, **kwargs):
        if (self._convolution is None) or (self._kwargs != kwargs):
            domain = self.bounding_box.domain(self._resolution)
            mesh = np.meshgrid(*domain)
            data = super().__call__(*mesh, **kwargs)

            from scipy.interpolate import RegularGridInterpolator
            convolution = RegularGridInterpolator(domain, data)

            if self._cache_convolution:
                self._kwargs = kwargs
                self._convolution = convolution

        else:
            convolution = self._convolution

        return convolution

    @staticmethod
    def _convolution_inputs(*args):
        not_scalar = np.where([not np.isscalar(arg) for arg in args])[0]

        if len(not_scalar) == 0:
            return np.array(args), (1,)
        else:
            output_shape = args[not_scalar[0]].shape
            if not all(args[index].shape == output_shape for index in not_scalar):
                raise ValueError('Values have differing shapes')

            inputs = []
            for arg in args:
                if np.isscalar(arg):
                    inputs.append(np.full(output_shape, arg))
                else:
                    inputs.append(arg)

            return np.reshape(inputs, (len(inputs), -1)).T, output_shape

    @staticmethod
    def _convolution_outputs(outputs, output_shape):
        return outputs.reshape(output_shape)

    def __call__(self, *args, **kw):
        inputs, output_shape = self._convolution_inputs(*args)
        convolution = self._get_convolution(**kw)
        outputs = convolution(inputs)

        return self._convolution_outputs(outputs, output_shape)
